# SparkSQL does not support 'order by' for its 'listagg' equivalent
# the argument is ignored, so let's ignore those fields when checking equivalency

models__test_listagg_no_order_by_sql = """
with data as (
    select * from {{ ref('data_listagg') }}
),
data_output as (
    select * from {{ ref('data_listagg_output') }}
),
calculate as (
/*

    select
        group_col,
        {{ listagg('string_text', "'_|_'", "order by order_col") }} as actual,
        'bottom_ordered' as version
    from data
    group by group_col
    union all
    select
        group_col,
        {{ listagg('string_text', "'_|_'", "order by order_col", 2) }} as actual,
        'bottom_ordered_limited' as version
    from data
    group by group_col
    union all

*/
    select
        group_col,
        {{ listagg('string_text', "', '") }} as actual,
        'comma_whitespace_unordered' as version
    from data
    where group_col = 3
    group by group_col
    union all
    select
        group_col,
        {{ listagg('DISTINCT string_text', "','") }} as actual,
        'distinct_comma' as version
    from data
    where group_col = 3
    group by group_col
    union all
    select
        group_col,
        {{ listagg('string_text') }} as actual,
        'no_params' as version
    from data
    where group_col = 3
    group by group_col
)
select
    calculate.actual,
    data_output.expected
from calculate
left join data_output
on calculate.group_col = data_output.group_col
and calculate.version = data_output.version
"""



import pytest

from dbt.tests.util import run_dbt, check_relations_equal
from dbt.tests.adapter.simple_seed.test_seed import SeedConfigBase
from tests.functional.adapter.incremental_strategies.seeds import (
    expected_append_csv,
    expected_overwrite_csv,
    expected_upsert_csv,
    expected_partial_upsert_csv,
)
from tests.functional.adapter.incremental_strategies.fixtures import (
    bad_file_format_sql,
    bad_merge_not_delta_sql,
    bad_strategy_sql,
    default_append_sql,
    insert_overwrite_no_partitions_sql,
    insert_overwrite_partitions_sql,
    append_delta_sql,
    delta_merge_no_key_sql,
    delta_merge_unique_key_sql,
    delta_merge_update_columns_sql,
    # Skip: CT-1873 insert_overwrite_partitions_delta_sql,
)


class BaseIncrementalStrategies(SeedConfigBase):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "expected_append.csv": expected_append_csv,
            "expected_overwrite.csv": expected_overwrite_csv,
            "expected_upsert.csv": expected_upsert_csv,
            "expected_partial_upsert.csv": expected_partial_upsert_csv,
        }

    @staticmethod
    def seed_and_run_once():
        run_dbt(["seed"])
        run_dbt(["run"])

    @staticmethod
    def seed_and_run_twice():
        run_dbt(["seed"])
        run_dbt(["run"])
        run_dbt(["run"])


class TestDefaultAppend(BaseIncrementalStrategies):
    @pytest.fixture(scope="class")
    def models(self):
        return {"default_append.sql": default_append_sql}

    def run_and_test(self, project):
        self.seed_and_run_twice()
        check_relations_equal(project.adapter, ["default_append", "expected_append"])

    @pytest.mark.skip_profile(
        "databricks_http_cluster", "databricks_sql_endpoint", "spark_session", "spark_http_odbc"
    )
    def test_default_append(self, project):
        self.run_and_test(project)


class TestInsertOverwrite(BaseIncrementalStrategies):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "insert_overwrite_no_partitions.sql": insert_overwrite_no_partitions_sql,
            "insert_overwrite_partitions.sql": insert_overwrite_partitions_sql,
        }

    def run_and_test(self, project):
        self.seed_and_run_twice()
        check_relations_equal(
            project.adapter, ["insert_overwrite_no_partitions", "expected_overwrite"]
        )
        check_relations_equal(project.adapter, ["insert_overwrite_partitions", "expected_upsert"])

    @pytest.mark.skip_profile(
        "databricks_http_cluster", "databricks_sql_endpoint", "spark_session", "spark_http_odbc"
    )
    def test_insert_overwrite(self, project):
        self.run_and_test(project)


class TestDeltaStrategies(BaseIncrementalStrategies):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "append_delta.sql": append_delta_sql,
            "merge_no_key.sql": delta_merge_no_key_sql,
            "merge_unique_key.sql": delta_merge_unique_key_sql,
            "merge_update_columns.sql": delta_merge_update_columns_sql,
            # Skip: cannot be acnive on any endpoint with grants
            # "insert_overwrite_partitions_delta.sql": insert_overwrite_partitions_delta_sql,
        }

    def run_and_test(self, project):
        self.seed_and_run_twice()
        check_relations_equal(project.adapter, ["append_delta", "expected_append"])
        check_relations_equal(project.adapter, ["merge_no_key", "expected_append"])
        check_relations_equal(project.adapter, ["merge_unique_key", "expected_upsert"])
        check_relations_equal(project.adapter, ["merge_update_columns", "expected_partial_upsert"])

    @pytest.mark.skip_profile(
        "apache_spark",
        "databricks_http_cluster",
        "databricks_sql_endpoint",
        "spark_session",
        "spark_http_odbc",
    )
    def test_delta_strategies(self, project):
        self.run_and_test(project)

    @pytest.mark.skip(
        reason="this feature is incompatible with databricks settings required for grants"
    )
    def test_delta_strategies_overwrite(self, project):
        self.seed_and_run_twice()
        check_relations_equal(
            project.adapter, ["insert_overwrite_partitions_delta", "expected_upsert"]
        )


class TestBadStrategies(BaseIncrementalStrategies):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "bad_file_format.sql": bad_file_format_sql,
            "bad_merge_not_delta.sql": bad_merge_not_delta_sql,
            "bad_strategy.sql": bad_strategy_sql,
        }

    @staticmethod
    def run_and_test():
        run_results = run_dbt(["run"], expect_pass=False)
        # assert all models fail with compilation errors
        for result in run_results:
            assert result.status == "error"
            assert "Compilation Error in model" in result.message

    @pytest.mark.skip_profile("databricks_http_cluster", "spark_session")
    def test_bad_strategies(self, project):
        self.run_and_test()



"""Functional tests for query timeout and retry behavior.

NOTE: These tests only work with PyHive-based connections (http/thrift methods).
ODBC connections use PyodbcConnectionWrapper which doesn't have timeout/retry support yet.
"""

import pytest
from dbt.tests.util import run_dbt, run_dbt_and_capture
from dbt_common.exceptions import DbtRuntimeError


# Model that creates a delay to simulate a long-running query
# Uses a large cross join to create computational delay
long_running_model = """
{{ config(materialized='table') }}
-- Generate a large dataset to create a delay
-- This creates roughly 1 million rows which should take several seconds
with numbers as (
  select explode(sequence(1, 100000)) as n
),
cross_product as (
  select a.n as n1, b.n as n2
  from numbers a
  cross join numbers b
)
select count(*) as total_count
from cross_product
"""

simple_model = """
{{ config(materialized='table') }}
select 1 as id
"""


@pytest.mark.skip_profile(
    "spark_http_odbc",
    "databricks_cluster",
    "databricks_sql_endpoint",
    "databricks_http_cluster",
    "spark_session",
)
class TestQueryTimeout:
    """Test query timeout functionality.

    Skipped on ODBC profiles because PyodbcConnectionWrapper doesn't use
    the async polling mechanism with timeout support.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "long_running.sql": long_running_model,
            "simple.sql": simple_model,
        }

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target):
        """Override profile to add timeout configuration."""
        dbt_profile_target["query_timeout"] = 1  # 1 second timeout
        dbt_profile_target["poll_interval"] = 1  # Poll every second for faster test
        dbt_profile_target["query_retries"] = 0  # Disable retries for clearer errors
        return dbt_profile_target

    def test_query_timeout_exceeded(self, project):
        """Test that queries exceeding timeout raise appropriate error."""

        # Simple model should succeed (runs quickly)
        results = run_dbt(["run", "--select", "simple"])
        assert results[0].status == "success"

        # Long-running model should timeout
        # The cross-join query should take longer than 2 seconds on most systems
        _, output = run_dbt_and_capture(["run", "--select", "long_running"], expect_pass=False)

        assert "exceeded timeout" in output.lower()



import pytest
from dbt.tests.adapter.dbt_clone.test_dbt_clone import BaseClonePossible
from tests.functional.adapter.dbt_clone.fixtures import (
    seed_csv,
    table_model_sql,
    view_model_sql,
    ephemeral_model_sql,
    exposures_yml,
    schema_yml,
    snapshot_sql,
    get_schema_name_sql,
    macros_sql,
    infinite_macros_sql,
)


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestSparkClonePossible(BaseClonePossible):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_model.sql": table_model_sql,
            "view_model.sql": view_model_sql,
            "ephemeral_model.sql": ephemeral_model_sql,
            "schema.yml": schema_yml,
            "exposures.yml": exposures_yml,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "macros.sql": macros_sql,
            "infinite_macros.sql": infinite_macros_sql,
            "get_schema_name.sql": get_schema_name_sql,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seed_csv,
        }

    @pytest.fixture(scope="class")
    def snapshots(self):
        return {
            "snapshot.sql": snapshot_sql,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
            },
            "seeds": {
                "test": {
                    "quote_columns": False,
                },
                "+file_format": "delta",
            },
            "snapshots": {
                "+file_format": "delta",
            },
        }

    @pytest.fixture(autouse=True)
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=f"{project.test_schema}_seeds"
            )
            project.adapter.drop_schema(relation)

            relation = project.adapter.Relation.create(
                database=project.database, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    pass



import pytest

from dbt.tests.util import run_dbt

from fixtures import (
    _MODELS__MY_FUN_DOCS,
    _MODELS__INCREMENTAL_DELTA,
    _MODELS__TABLE_DELTA_MODEL,
    _MODELS__TABLE_DELTA_MODEL_MISSING_COLUMN,
    _PROPERTIES__MODELS,
    _PROPERTIES__SEEDS,
    _SEEDS__BASIC,
    _MODELS__VIEW_DELTA_MODEL,
    _VIEW_PROPERTIES_MODELS,
)


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestPersistDocsDeltaTable:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "incremental_delta_model.sql": _MODELS__INCREMENTAL_DELTA,
            "my_fun_docs.md": _MODELS__MY_FUN_DOCS,
            "table_delta_model.sql": _MODELS__TABLE_DELTA_MODEL,
            "schema.yml": _PROPERTIES__MODELS,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": _SEEDS__BASIC, "seed.yml": _PROPERTIES__SEEDS}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "test": {
                    "+persist_docs": {
                        "relation": True,
                        "columns": True,
                    },
                }
            },
            "seeds": {
                "test": {
                    "+persist_docs": {
                        "relation": True,
                        "columns": True,
                    },
                    "+file_format": "delta",
                    "+quote_columns": True,
                }
            },
        }

    def test_delta_comments(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])

        for table, whatis in [
            ("table_delta_model", "Table"),
            ("seed", "Seed"),
            ("incremental_delta_model", "Incremental"),
        ]:
            results = project.run_sql(
                "describe extended {schema}.{table}".format(
                    schema=project.test_schema, table=table
                ),
                fetch="all",
            )

            for result in results:
                if result[0] == "Comment":
                    assert result[1].startswith(f"{whatis} model description")
                if result[0] == "id":
                    assert result[2].startswith("id Column description")
                if result[0] == "name":
                    assert result[2].startswith("Some stuff here and then a call to")


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestPersistDocsDeltaView:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_delta_model.sql": _MODELS__TABLE_DELTA_MODEL,
            "view_delta_model.sql": _MODELS__VIEW_DELTA_MODEL,
            "schema.yml": _VIEW_PROPERTIES_MODELS,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "test": {
                    "+persist_docs": {
                        "relation": True,
                        "columns": True,
                    },
                }
            },
        }

    def test_delta_comments(self, project):
        run_dbt(["run"])

        results = project.run_sql(
            "describe extended {schema}.{table}".format(
                schema=project.test_schema, table="view_delta_model"
            ),
            fetch="all",
        )

        for result in results:
            if result[0] == "Comment":
                assert result[1].startswith("View model description")
            if result[0] == "id":
                assert result[2].startswith("id Column description")
            if result[0] == "count":
                assert result[2] is None


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestPersistDocsMissingColumn:
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "test": {
                    "+persist_docs": {
                        "columns": True,
                    },
                }
            }
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": _SEEDS__BASIC, "seed.yml": _PROPERTIES__SEEDS}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_delta_model.sql": _MODELS__TABLE_DELTA_MODEL_MISSING_COLUMN,
            "my_fun_docs.md": _MODELS__MY_FUN_DOCS,
        }

    @pytest.fixture(scope="class")
    def properties(self):
        return {"schema.yml": _PROPERTIES__MODELS}

    def test_missing_column(self, project):
        """
        spark will use our schema to verify all columns exist rather than fail silently

        "resolve" vs "resolved" is intentional; example error message:
            ('42000', '[42000] [Simba][Hardy] (80) Syntax or semantic analysis error thrown in server while executing query.
            Error message from server:
                org.apache.hive.service.cli.HiveSQLException: Error running query: [UNRESOLVED_COLUMN.WITH_SUGGESTION]
                org.apache.spark.sql.AnalysisException: [UNRESOLVED_COLUMN.WITH_SUGGESTION] A column, variable, or function parameter with name `name` cannot be resolve (80) (SQLExecDirectW)'
            )
        """
        run_dbt(["seed"])
        res = run_dbt(["run"], expect_pass=False)
        assert any(
            [
                "[UNRESOLVED_COLUMN.WITH_SUGGESTION]" in res[0].message,
                "Missing field name in table" in res[0].message,
            ]
        )



import pytest
from dbt.tests.adapter.incremental.test_incremental_predicates import BaseIncrementalPredicates


models__spark_incremental_predicates_sql = """
{{ config(
    materialized = 'incremental',
    unique_key = 'id'
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg, 'blue' as color
union all
select cast(2 as bigint) as id, 'goodbye' as msg, 'red' as color

{% else %}

-- merge will not happen on the above record where id = 2, so new record will fall to insert
select cast(1 as bigint) as id, 'hey' as msg, 'blue' as color
union all
select cast(2 as bigint) as id, 'yo' as msg, 'green' as color
union all
select cast(3 as bigint) as id, 'anyway' as msg, 'purple' as color

{% endif %}
"""


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestIncrementalPredicatesMergeSpark(BaseIncrementalPredicates):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+incremental_predicates": ["dbt_internal_dest.id != 2"],
                "+incremental_strategy": "merge",
                "+file_format": "delta",
            }
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "delete_insert_incremental_predicates.sql": models__spark_incremental_predicates_sql
        }


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestPredicatesMergeSpark(BaseIncrementalPredicates):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+predicates": ["dbt_internal_dest.id != 2"],
                "+incremental_strategy": "merge",
                "+file_format": "delta",
            }
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "delete_insert_incremental_predicates.sql": models__spark_incremental_predicates_sql
        }



import pytest

from tests.unit.utils import config_from_parts_or_dicts


@pytest.fixture(scope="session", autouse=True)
def base_project_cfg():
    return {
        "name": "X",
        "version": "0.1",
        "profile": "test",
        "project-root": "/tmp/dbt/does-not-exist",
        "quoting": {
            "identifier": False,
            "schema": False,
        },
        "config-version": 2,
    }


@pytest.fixture(scope="session", autouse=True)
def target_http(base_project_cfg):
    config = config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "http",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 443,
                    "token": "abc123",
                    "organization": "0123456789",
                    "cluster": "01234-23423-coffeetime",
                    "server_side_parameters": {"spark.driver.memory": "4g"},
                }
            },
            "target": "test",
        },
    )
    return config


@pytest.fixture(scope="session", autouse=True)
def target_thrift(base_project_cfg):
    return config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "thrift",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 10001,
                    "user": "dbt",
                }
            },
            "target": "test",
        },
    )


@pytest.fixture(scope="session", autouse=True)
def target_thrift_kerberos(base_project_cfg):
    return config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "thrift",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 10001,
                    "user": "dbt",
                    "auth": "KERBEROS",
                    "kerberos_service_name": "hive",
                }
            },
            "target": "test",
        },
    )


@pytest.fixture(scope="session", autouse=True)
def target_use_ssl_thrift(base_project_cfg):
    return config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "thrift",
                    "use_ssl": True,
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 10001,
                    "user": "dbt",
                }
            },
            "target": "test",
        },
    )


@pytest.fixture(scope="session", autouse=True)
def target_odbc_cluster(base_project_cfg):
    return config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "odbc",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 443,
                    "token": "abc123",
                    "organization": "0123456789",
                    "cluster": "01234-23423-coffeetime",
                    "driver": "Simba",
                }
            },
            "target": "test",
        },
    )


@pytest.fixture(scope="session", autouse=True)
def target_odbc_sql_endpoint(base_project_cfg):
    return config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "odbc",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 443,
                    "token": "abc123",
                    "endpoint": "012342342393920a",
                    "driver": "Simba",
                }
            },
            "target": "test",
        },
    )


@pytest.fixture(scope="session", autouse=True)
def target_odbc_with_extra_conn(base_project_cfg):
    return config_from_parts_or_dicts(
        base_project_cfg,
        {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "odbc",
                    "host": "myorg.sparkhost.com",
                    "schema": "analytics",
                    "port": 443,
                    "driver": "Simba",
                    "connection_string_suffix": "someExtraValues",
                    "connect_retries": 3,
                    "connect_timeout": 5,
                    "retry_all": True,
                }
            },
            "target": "test",
        },
    )



import pytest

from dbt.tests.adapter.dbt_show.test_dbt_show import (
    BaseShowSqlHeader,
    BaseShowLimit,
    BaseShowDoesNotHandleDoubleLimit,
)


class TestSparkShowLimit(BaseShowLimit):
    pass


class TestSparkShowSqlHeader(BaseShowSqlHeader):
    pass


@pytest.mark.skip_profile("apache_spark", "spark_session", "databricks_http_cluster")
class TestSparkShowDoesNotHandleDoubleLimit(BaseShowDoesNotHandleDoubleLimit):
    """The syntax message is quite variable across clusters, but this hits two at once."""

    DATABASE_ERROR_MESSAGE = "limit"



import pytest
from dbt.tests.util import run_dbt
from tests.functional.adapter.seed_column_types.fixtures import (
    _MACRO_TEST_IS_TYPE_SQL,
    _SEED_CSV,
    _SEED_YML,
)


@pytest.mark.skip_profile("spark_session")
class TestSeedColumnTypesCast:
    @pytest.fixture(scope="class")
    def macros(self):
        return {"test_is_type.sql": _MACRO_TEST_IS_TYPE_SQL}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"payments.csv": _SEED_CSV, "schema.yml": _SEED_YML}

    #  We want to test seed types because hive would cause all fields to be strings.
    # setting column_types in project.yml should change them and pass.
    def test_column_seed_type(self, project):
        results = run_dbt(["seed"])
        assert len(results) == 1
        run_dbt(["test"], expect_pass=False)



import os
import pytest
from dbt.tests.util import run_dbt, write_file
from dbt.tests.adapter.python_model.test_python_model import (
    BasePythonModelTests,
    BasePythonIncrementalTests,
)
from dbt.tests.adapter.python_model.test_spark import BasePySparkTests


@pytest.mark.skip_profile(
    "apache_spark", "spark_session", "databricks_sql_endpoint", "spark_http_odbc"
)
class TestPythonModelSpark(BasePythonModelTests):
    pass


@pytest.mark.skip_profile(
    "apache_spark", "spark_session", "databricks_sql_endpoint", "spark_http_odbc"
)
class TestPySpark(BasePySparkTests):
    def test_different_dataframes(self, project):
        """
        Test that python models are supported using dataframes from:
        - pandas
        - pyspark
        - pyspark.pandas (formerly dataspark.koalas)

        Note:
            The CI environment is on Apache Spark >3.1, which includes koalas as pyspark.pandas.
            The only Databricks runtime that supports Apache Spark <=3.1 is 9.1 LTS, which is EOL 2024-09-23.
            For more information, see:
            - https://github.com/databricks/koalas
            - https://docs.databricks.com/en/release-notes/runtime/index.html
        """
        results = run_dbt(["run", "--exclude", "koalas_df"])
        assert len(results) == 3


@pytest.mark.skip_profile(
    "apache_spark", "spark_session", "databricks_sql_endpoint", "spark_http_odbc"
)
class TestPythonIncrementalModelSpark(BasePythonIncrementalTests):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {}


models__simple_python_model = """
def model(dbt, spark):
    dbt.config(
        materialized='table',
        submission_method='job_cluster',
        job_cluster_config={
            "spark_version": "12.2.x-scala2.12",
            "node_type_id": "i3.xlarge",
            "num_workers": 0,
            "spark_conf": {
                "spark.databricks.cluster.profile": "singleNode",
                "spark.master": "local[*, 4]"
            },
            "custom_tags": {
                "ResourceClass": "SingleNode"
            }
        },
        packages=['pydantic']
    )
    data = [[1,2]] * 10
    return spark.createDataFrame(data, schema=['test', 'test2'])
"""
models__simple_python_model_v2 = """
import pandas

def model(dbt, spark):
    dbt.config(
        materialized='table',
    )
    data = [[1,2]] * 10
    return spark.createDataFrame(data, schema=['test1', 'test3'])
"""


@pytest.mark.skip_profile(
    "apache_spark",
    "spark_session",
    "databricks_sql_endpoint",
    "spark_http_odbc",
    "databricks_http_cluster",
)
class TestChangingSchemaSpark:
    """
    Confirm that we can setup a spot instance and parse required packages into the Databricks job.

    Notes:
        - This test generates a spot instance on demand using the settings from `job_cluster_config`
        in `models__simple_python_model` above. It takes several minutes to run due to creating the cluster.
        The job can be monitored via "Data Engineering > Job Runs" or "Workflows > Job Runs"
        in the Databricks UI (instead of via the normal cluster).
        - The `spark_version` argument will need to periodically be updated. It will eventually become
        unsupported and start experiencing issues.
        - See https://github.com/explosion/spaCy/issues/12659 for why we're pinning pydantic
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {"simple_python_model.py": models__simple_python_model}

    def test_changing_schema_with_log_validation(self, project, logs_dir):
        run_dbt(["run"])
        write_file(
            models__simple_python_model_v2,
            project.project_root + "/models",
            "simple_python_model.py",
        )
        run_dbt(["run"])
        log_file = os.path.join(logs_dir, "dbt.log")
        with open(log_file, "r") as f:
            log = f.read()
            # validate #5510 log_code_execution works
            assert "On model.test.simple_python_model:" in log
            assert "spark.createDataFrame(data, schema=['test1', 'test3'])" in log
            assert "Execution status: OK in" in log



import pytest
from dbt.tests.adapter.grants.test_model_grants import BaseModelGrants
from dbt.tests.adapter.grants.test_incremental_grants import BaseIncrementalGrants
from dbt.tests.adapter.grants.test_invalid_grants import BaseInvalidGrants
from dbt.tests.adapter.grants.test_seed_grants import BaseSeedGrants
from dbt.tests.adapter.grants.test_snapshot_grants import BaseSnapshotGrants


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestModelGrantsSpark(BaseModelGrants):
    def privilege_grantee_name_overrides(self):
        # insert --> modify
        return {
            "select": "select",
            "insert": "modify",
            "fake_privilege": "fake_privilege",
            "invalid_user": "invalid_user",
        }


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestIncrementalGrantsSpark(BaseIncrementalGrants):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
                "+incremental_strategy": "merge",
            }
        }


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestSeedGrantsSpark(BaseSeedGrants):
    # seeds in dbt-spark are currently "full refreshed," in such a way that
    # the grants are not carried over
    # see https://github.com/dbt-labs/dbt-spark/issues/388
    def seeds_support_partial_refresh(self):
        return False


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestSnapshotGrantsSpark(BaseSnapshotGrants):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "snapshots": {
                "+file_format": "delta",
                "+incremental_strategy": "merge",
            }
        }


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestInvalidGrantsSpark(BaseInvalidGrants):
    def grantee_does_not_exist_error(self):
        return "RESOURCE_DOES_NOT_EXIST"

    def privilege_does_not_exist_error(self):
        return "Action Unknown"



import unittest

from dbt.adapters.spark import SparkColumn


class TestSparkColumn(unittest.TestCase):
    def test_convert_table_stats_with_no_statistics(self):
        self.assertDictEqual(SparkColumn.convert_table_stats(None), {})

    def test_convert_table_stats_with_bytes(self):
        self.assertDictEqual(
            SparkColumn.convert_table_stats("123456789 bytes"),
            {
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 123456789,
            },
        )

    def test_convert_table_stats_with_bytes_and_rows(self):
        self.assertDictEqual(
            SparkColumn.convert_table_stats("1234567890 bytes, 12345678 rows"),
            {
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 1234567890,
                "stats:rows:description": "",
                "stats:rows:include": True,
                "stats:rows:label": "rows",
                "stats:rows:value": 12345678,
            },
        )



_MACRO_TEST_IS_TYPE_SQL = """
{% macro simple_type_check_column(column, check) %}
    {% set checks = {
        'string': column.is_string,
        'float': column.is_float,
        'number': column.is_number,
        'numeric': column.is_numeric,
        'integer': column.is_integer,
    } %}

    {{ return(checks[check]()) }}
{% endmacro %}

{% macro type_check_column(column, type_checks) %}
    {% set failures = [] %}
    {% for type_check in type_checks %}
        {% if type_check.startswith('not ') %}
            {% if simple_type_check_column(column, type_check[4:]) %}
                {% do failures.append(type_check) %}
            {% endif %}
        {% else %}
            {% if not simple_type_check_column(column, type_check) %}
                {% do failures.append(type_check) %}
            {% endif %}
        {% endif %}
    {% endfor %}

    {% do return((failures | length) == 0) %}
{% endmacro %}

{% macro is_bad_column(column, column_map) %}
    {% set column_key = (column.name | lower) %}
    {% if column_key not in column_map %}
        {% do exceptions.raise_compiler_error('column key ' ~ column_key ~ ' not found in ' ~ (column_map | list | string)) %}
    {% endif %}

    {% set type_checks = column_map[column_key] %}
    {% if not type_checks %}
        {% do exceptions.raise_compiler_error('no type checks?') %}
    {% endif %}

    {{ return(not type_check_column(column, type_checks)) }}
{% endmacro %}

{% test is_type(model, column_map) %}
    {% if not execute %}
        {{ return(None) }}
    {% endif %}

    {% set columns = adapter.get_columns_in_relation(model) %}
    {% if (column_map | length) != (columns | length) %}
        {% set column_map_keys = (column_map | list | string) %}
        {% set column_names = (columns | map(attribute='name') | list | string) %}
        {% do exceptions.raise_compiler_error('did not get all the columns/all columns not specified:\n' ~ column_map_keys ~ '\nvs\n' ~ column_names) %}
    {% endif %}

    {% set bad_columns = [] %}
    {% for column in columns %}
        {% if is_bad_column(column, column_map) %}
            {% do bad_columns.append(column.name) %}
        {% endif %}
    {% endfor %}

    {% set num_bad_columns = (bad_columns | length) %}

    select '{{ num_bad_columns }}' as bad_column
    group by 1
    having bad_column > 0

{% endtest %}
""".strip()


_SEED_CSV = """
id,orderid,paymentmethod,status,amount,amount_usd,created
1,1,credit_card,success,1000,10.00,2018-01-01
2,2,credit_card,success,2000,20.00,2018-01-02
3,3,coupon,success,100,1.00,2018-01-04
4,4,coupon,success,2500,25.00,2018-01-05
5,5,bank_transfer,fail,1700,17.00,2018-01-05
6,5,bank_transfer,success,1700,17.00,2018-01-05
7,6,credit_card,success,600,6.00,2018-01-07
8,7,credit_card,success,1600,16.00,2018-01-09
9,8,credit_card,success,2300,23.00,2018-01-11
10,9,gift_card,success,2300,23.00,2018-01-12
""".strip()


_SEED_YML = """
version: 2

seeds:
  - name: payments
    config:
        column_types:
            id: string
            orderid: string
            paymentmethod: string
            status: string
            amount: integer
            amount_usd: decimal(20,2)
            created: timestamp
    tests:
        - is_type:
            column_map:
                id: ["string", "not number"]
                orderid: ["string", "not number"]
                paymentmethod: ["string", "not number"]
                status: ["string", "not number"]
                amount: ["integer", "number"]
                amount_usd: ["decimal", "number"]
                created: ["timestamp", "string"]
""".strip()



import pytest

from dbt.tests.util import run_dbt, relation_from_name, check_relations_equal_with_relations


_MODEL_CHILD = "select 1"


_MODEL_PARENT = """
{% set cols = adapter.get_columns_in_relation(ref('child')) %}

select
    {% for col in cols %}
        {{ adapter.quote(col.column) }}{%- if not loop.last %},{{ '\n ' }}{% endif %}
    {% endfor %}
from {{ ref('child') }}
"""


class TestColumnsInRelation:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "child.sql": _MODEL_CHILD,
            "parent.sql": _MODEL_PARENT,
        }

    @pytest.mark.skip_profile("databricks_http_cluster", "spark_session")
    def test_get_columns_in_relation(self, project):
        run_dbt(["run"])
        child = relation_from_name(project.adapter, "child")
        parent = relation_from_name(project.adapter, "parent")
        check_relations_equal_with_relations(project.adapter, [child, parent])



import pytest

from dbt.tests.adapter.utils.test_array_append import BaseArrayAppend
from dbt.tests.adapter.utils.test_array_concat import BaseArrayConcat
from dbt.tests.adapter.utils.test_array_construct import BaseArrayConstruct
from dbt.tests.adapter.utils.test_any_value import BaseAnyValue
from dbt.tests.adapter.utils.test_bool_or import BaseBoolOr
from dbt.tests.adapter.utils.test_cast import BaseCast
from dbt.tests.adapter.utils.test_cast_bool_to_text import BaseCastBoolToText
from dbt.tests.adapter.utils.test_concat import BaseConcat
from dbt.tests.adapter.utils.test_current_timestamp import BaseCurrentTimestampNaive
from dbt.tests.adapter.utils.test_date import BaseDate
from dbt.tests.adapter.utils.test_dateadd import BaseDateAdd
from dbt.tests.adapter.utils.test_datediff import BaseDateDiff
from dbt.tests.adapter.utils.test_date_trunc import BaseDateTrunc
from dbt.tests.adapter.utils.test_escape_single_quotes import BaseEscapeSingleQuotesBackslash
from dbt.tests.adapter.utils.test_except import BaseExcept
from dbt.tests.adapter.utils.test_hash import BaseHash
from dbt.tests.adapter.utils.test_intersect import BaseIntersect
from dbt.tests.adapter.utils.test_last_day import BaseLastDay
from dbt.tests.adapter.utils.test_length import BaseLength
from dbt.tests.adapter.utils.test_position import BasePosition
from dbt.tests.adapter.utils.test_replace import BaseReplace
from dbt.tests.adapter.utils.test_right import BaseRight
from dbt.tests.adapter.utils.test_safe_cast import BaseSafeCast

from dbt.tests.adapter.utils.test_split_part import BaseSplitPart
from dbt.tests.adapter.utils.test_string_literal import BaseStringLiteral

# requires modification
from dbt.tests.adapter.utils.test_listagg import BaseListagg
from dbt.tests.adapter.utils.fixture_listagg import models__test_listagg_yml
from tests.functional.adapter.utils.fixture_listagg import models__test_listagg_no_order_by_sql

seeds__data_split_part_csv = """parts,split_on,result_1,result_2,result_3,result_4
a|b|c,|,a,b,c,c
1|2|3,|,1,2,3,3
EMPTY|EMPTY|EMPTY,|,EMPTY,EMPTY,EMPTY,EMPTY
"""

seeds__data_last_day_csv = """date_day,date_part,result
2018-01-02,month,2018-01-31
2018-01-02,quarter,2018-03-31
2018-01-02,year,2018-12-31
"""


# skipped: ,month,


class TestAnyValue(BaseAnyValue):
    pass


class TestArrayAppend(BaseArrayAppend):
    pass


class TestArrayConcat(BaseArrayConcat):
    pass


class TestArrayConstruct(BaseArrayConstruct):
    pass


class TestBoolOr(BaseBoolOr):
    pass


class TestCast(BaseCast):
    pass


class TestCastBoolToText(BaseCastBoolToText):
    pass


@pytest.mark.skip_profile("spark_session")
class TestConcat(BaseConcat):
    pass


# Use either BaseCurrentTimestampAware or BaseCurrentTimestampNaive but not both
class TestCurrentTimestamp(BaseCurrentTimestampNaive):
    pass


class TestDate(BaseDate):
    pass


class TestDateAdd(BaseDateAdd):
    pass


# this generates too much SQL to run successfully in our testing environments :(
@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestDateDiff(BaseDateDiff):
    pass


class TestDateTrunc(BaseDateTrunc):
    pass


class TestEscapeSingleQuotes(BaseEscapeSingleQuotesBackslash):
    pass


class TestExcept(BaseExcept):
    pass


@pytest.mark.skip_profile("spark_session")
class TestHash(BaseHash):
    pass


class TestIntersect(BaseIntersect):
    pass


@pytest.mark.skip_profile("spark_session")  # spark session crashes in CI
class TestLastDay(BaseLastDay):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_last_day.csv": seeds__data_last_day_csv}


class TestLength(BaseLength):
    pass


# SparkSQL does not support 'order by' for its 'listagg' equivalent
# the argument is ignored, so let's ignore those fields when checking equivalency
class TestListagg(BaseListagg):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_listagg.yml": models__test_listagg_yml,
            "test_listagg.sql": self.interpolate_macro_namespace(
                models__test_listagg_no_order_by_sql, "listagg"
            ),
        }


class TestPosition(BasePosition):
    pass


@pytest.mark.skip_profile("spark_session")
class TestReplace(BaseReplace):
    pass


@pytest.mark.skip_profile("spark_session")
class TestRight(BaseRight):
    pass


class TestSafeCast(BaseSafeCast):
    pass


class TestSplitPart(BaseSplitPart):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_split_part.csv": seeds__data_split_part_csv}


class TestStringLiteral(BaseStringLiteral):
    pass



import pytest

from dbt.tests.adapter.basic.test_base import BaseSimpleMaterializations
from dbt.tests.adapter.basic.test_singular_tests import BaseSingularTests
from dbt.tests.adapter.basic.test_singular_tests_ephemeral import (
    BaseSingularTestsEphemeral,
)
from dbt.tests.adapter.basic.test_empty import BaseEmpty
from dbt.tests.adapter.basic.test_ephemeral import BaseEphemeral
from dbt.tests.adapter.basic.test_incremental import BaseIncremental
from dbt.tests.adapter.basic.test_generic_tests import BaseGenericTests
from dbt.tests.adapter.basic.test_snapshot_check_cols import BaseSnapshotCheckCols
from dbt.tests.adapter.basic.test_snapshot_timestamp import BaseSnapshotTimestamp
from dbt.tests.adapter.basic.test_adapter_methods import BaseAdapterMethod


@pytest.mark.skip_profile("spark_session")
class TestSimpleMaterializationsSpark(BaseSimpleMaterializations):
    pass


class TestSingularTestsSpark(BaseSingularTests):
    pass


# The local cluster currently tests on spark 2.x, which does not support this
# if we upgrade it to 3.x, we can enable this test
@pytest.mark.skip_profile("apache_spark")
class TestSingularTestsEphemeralSpark(BaseSingularTestsEphemeral):
    pass


class TestEmptySpark(BaseEmpty):
    pass


@pytest.mark.skip_profile("spark_session")
class TestEphemeralSpark(BaseEphemeral):
    pass


@pytest.mark.skip_profile("spark_session")
class TestIncrementalSpark(BaseIncremental):
    pass


class TestGenericTestsSpark(BaseGenericTests):
    pass


# These tests were not enabled in the dbtspec files, so skipping here.
# Error encountered was: Error running query: java.lang.ClassNotFoundException: delta.DefaultSource
@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestSnapshotCheckColsSpark(BaseSnapshotCheckCols):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seeds": {
                "+file_format": "delta",
            },
            "snapshots": {
                "+file_format": "delta",
            },
        }


# These tests were not enabled in the dbtspec files, so skipping here.
# Error encountered was: Error running query: java.lang.ClassNotFoundException: delta.DefaultSource
@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestSnapshotTimestampSpark(BaseSnapshotTimestamp):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seeds": {
                "+file_format": "delta",
            },
            "snapshots": {
                "+file_format": "delta",
            },
        }


@pytest.mark.skip_profile("spark_session")
class TestBaseAdapterMethod(BaseAdapterMethod):
    pass



#
# Models
#

default_append_sql = """
{{ config(
    materialized = 'incremental',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

#
# Bad Models
#

bad_file_format_sql = """
{{ config(
    materialized = 'incremental',
    file_format = 'something_else',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

bad_merge_not_delta_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

bad_strategy_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'something_else',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

#
# Delta Models
#

append_delta_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'append',
    file_format = 'delta',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

insert_overwrite_partitions_delta_sql = """
{{ config(
    materialized='incremental',
    incremental_strategy='insert_overwrite',
    partition_by='id',
    file_format='delta'
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
"""


delta_merge_no_key_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
    file_format = 'delta',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

delta_merge_unique_key_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
    file_format = 'delta',
    unique_key = 'id',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

delta_merge_update_columns_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
    file_format = 'delta',
    unique_key = 'id',
    merge_update_columns = ['msg'],
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg, 'blue' as color
union all
select cast(2 as bigint) as id, 'goodbye' as msg, 'red' as color

{% else %}

-- msg will be updated, color will be ignored
select cast(2 as bigint) as id, 'yo' as msg, 'green' as color
union all
select cast(3 as bigint) as id, 'anyway' as msg, 'purple' as color

{% endif %}
""".lstrip()

#
# Insert Overwrite
#

insert_overwrite_no_partitions_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'insert_overwrite',
    file_format = 'parquet',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

insert_overwrite_partitions_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'insert_overwrite',
    partition_by = 'id',
    file_format = 'parquet',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

#
# Hudi Models
#

append_hudi_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'append',
    file_format = 'hudi',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

hudi_insert_overwrite_no_partitions_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'insert_overwrite',
    file_format = 'hudi',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

hudi_insert_overwrite_partitions_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'insert_overwrite',
    partition_by = 'id',
    file_format = 'hudi',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

hudi_merge_no_key_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
    file_format = 'hudi',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

hudi_merge_unique_key_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
    file_format = 'hudi',
    unique_key = 'id',
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg
union all
select cast(2 as bigint) as id, 'goodbye' as msg

{% else %}

select cast(2 as bigint) as id, 'yo' as msg
union all
select cast(3 as bigint) as id, 'anyway' as msg

{% endif %}
""".lstrip()

hudi_update_columns_sql = """
{{ config(
    materialized = 'incremental',
    incremental_strategy = 'merge',
    file_format = 'hudi',
    unique_key = 'id',
    merge_update_columns = ['msg'],
) }}

{% if not is_incremental() %}

select cast(1 as bigint) as id, 'hello' as msg, 'blue' as color
union all
select cast(2 as bigint) as id, 'goodbye' as msg, 'red' as color

{% else %}

-- msg will be updated, color will be ignored
select cast(2 as bigint) as id, 'yo' as msg, 'green' as color
union all
select cast(3 as bigint) as id, 'anyway' as msg, 'purple' as color

{% endif %}
""".lstrip()



import unittest
from unittest import mock
import re
from jinja2 import Environment, FileSystemLoader


class TestSparkMacros(unittest.TestCase):
    def setUp(self):
        self.jinja_env = Environment(
            loader=FileSystemLoader("src/dbt/include/spark/macros"),
            extensions=[
                "jinja2.ext.do",
            ],
        )

        self.config = {}
        self.default_context = {
            "validation": mock.Mock(),
            "model": mock.Mock(),
            "exceptions": mock.Mock(),
            "config": mock.Mock(),
            "adapter": mock.Mock(),
            "return": lambda r: r,
        }
        self.default_context["config"].get = lambda key, default=None, **kwargs: self.config.get(
            key, default
        )

    def __get_template(self, template_filename):
        return self.jinja_env.get_template(template_filename, globals=self.default_context)

    def __run_macro(self, template, name, temporary, relation, sql):
        self.default_context["model"].alias = relation

        def dispatch(macro_name, macro_namespace=None, packages=None):
            return getattr(template.module, f"spark__{macro_name}")

        self.default_context["adapter"].dispatch = dispatch

        value = getattr(template.module, name)(temporary, relation, sql)
        return re.sub(r"\s\s+", " ", value)

    def test_macros_load(self):
        self.jinja_env.get_template("adapters.sql")

    def test_macros_create_table_as(self):
        template = self.__get_template("adapters.sql")
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()

        self.assertEqual(sql, "create table my_table as select 1")

    def test_macros_create_table_as_file_format(self):
        template = self.__get_template("adapters.sql")

        self.config["file_format"] = "delta"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(sql, "create or replace table my_table using delta as select 1")

        self.config["file_format"] = "hudi"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(sql, "create table my_table using hudi as select 1")

    def test_macros_create_table_as_options(self):
        template = self.__get_template("adapters.sql")

        self.config["file_format"] = "delta"
        self.config["options"] = {"compression": "gzip"}
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql,
            'create or replace table my_table using delta options (compression "gzip" ) as select 1',
        )

        self.config["file_format"] = "hudi"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql, 'create table my_table using hudi options (compression "gzip" ) as select 1'
        )

    def test_macros_create_table_as_hudi_options(self):
        template = self.__get_template("adapters.sql")

        self.config["file_format"] = "hudi"
        self.config["unique_key"] = "id"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1 as id"
        ).strip()
        self.assertEqual(
            sql, 'create table my_table using hudi options (primaryKey "id" ) as select 1 as id'
        )

        self.config["file_format"] = "hudi"
        self.config["unique_key"] = "id"
        self.config["options"] = {"primaryKey": "id"}
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1 as id"
        ).strip()
        self.assertEqual(
            sql, 'create table my_table using hudi options (primaryKey "id" ) as select 1 as id'
        )

        self.config["file_format"] = "hudi"
        self.config["unique_key"] = "uuid"
        self.config["options"] = {"primaryKey": "id"}
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1 as id"
        )
        self.assertIn("mock.raise_compiler_error()", sql)

    def test_macros_create_table_as_partition(self):
        template = self.__get_template("adapters.sql")

        self.config["partition_by"] = "partition_1"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(sql, "create table my_table partitioned by (partition_1) as select 1")

    def test_macros_create_table_as_partitions(self):
        template = self.__get_template("adapters.sql")

        self.config["partition_by"] = ["partition_1", "partition_2"]
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql, "create table my_table partitioned by (partition_1,partition_2) as select 1"
        )

    def test_macros_create_table_as_cluster(self):
        template = self.__get_template("adapters.sql")

        self.config["clustered_by"] = "cluster_1"
        self.config["buckets"] = "1"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql, "create table my_table clustered by (cluster_1) into 1 buckets as select 1"
        )

    def test_macros_create_table_as_clusters(self):
        template = self.__get_template("adapters.sql")

        self.config["clustered_by"] = ["cluster_1", "cluster_2"]
        self.config["buckets"] = "1"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql,
            "create table my_table clustered by (cluster_1,cluster_2) into 1 buckets as select 1",
        )

    def test_macros_create_table_as_location(self):
        template = self.__get_template("adapters.sql")

        self.config["location_root"] = "/mnt/root"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(sql, "create table my_table location '/mnt/root/my_table' as select 1")

    def test_macros_create_table_as_comment(self):
        template = self.__get_template("adapters.sql")

        self.config["persist_docs"] = {"relation": True}
        self.default_context["model"].description = "Description Test"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(sql, "create table my_table comment 'Description Test' as select 1")

    def test_macros_create_table_as_all(self):
        template = self.__get_template("adapters.sql")

        self.config["file_format"] = "delta"
        self.config["location_root"] = "/mnt/root"
        self.config["partition_by"] = ["partition_1", "partition_2"]
        self.config["clustered_by"] = ["cluster_1", "cluster_2"]
        self.config["buckets"] = "1"
        self.config["persist_docs"] = {"relation": True}
        self.default_context["model"].description = "Description Test"

        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql,
            "create or replace table my_table using delta partitioned by (partition_1,partition_2) clustered by (cluster_1,cluster_2) into 1 buckets location '/mnt/root/my_table' comment 'Description Test' as select 1",
        )

        self.config["file_format"] = "hudi"
        sql = self.__run_macro(
            template, "spark__create_table_as", False, "my_table", "select 1"
        ).strip()
        self.assertEqual(
            sql,
            "create table my_table using hudi partitioned by (partition_1,partition_2) clustered by (cluster_1,cluster_2) into 1 buckets location '/mnt/root/my_table' comment 'Description Test' as select 1",
        )



import os

import argparse
import sys
from typing import Dict, Tuple

import anyio as anyio
import dagger as dagger
from dotenv import find_dotenv, load_dotenv

PG_PORT = 5432
load_dotenv(find_dotenv("test.env"))
# if env vars aren't specified in test.env (i.e. in github actions worker), use the ones from the host
TESTING_ENV_VARS = {
    env_name: os.environ[env_name]
    for env_name in os.environ
    if env_name.startswith(("DD_", "DBT_"))
}

TESTING_ENV_VARS.update({"ODBC_DRIVER": "/opt/simba/spark/lib/64/libsparkodbc_sb64.so"})


def env_variables(envs: Dict[str, str]):
    def env_variables_inner(ctr: dagger.Container):
        for key, value in envs.items():
            ctr = ctr.with_env_variable(key, value)
        return ctr

    return env_variables_inner


def get_postgres_container(client: dagger.Client) -> Tuple[dagger.Container, str]:
    ctr = (
        client.container()
        .from_("postgres:13")
        .with_env_variable("POSTGRES_PASSWORD", "postgres")
        .with_exposed_port(PG_PORT)
        .as_service()
    )

    return ctr, "postgres_db"


def get_spark_container(client: dagger.Client) -> Tuple[dagger.Service, str]:
    spark_dir = client.host().directory("./dagger/spark-container")
    spark_ctr_base = (
        client.container()
        .from_("eclipse-temurin:8-jre")
        .with_directory("/spark_setup", spark_dir)
        .with_env_variable("SPARK_HOME", "/usr/spark")
        .with_env_variable("PATH", "/usr/spark/bin:/usr/spark/sbin:$PATH", expand=True)
        .with_file(
            "/scripts/entrypoint.sh",
            client.host().file("./dagger/spark-container/entrypoint.sh"),
            permissions=755,
        )
        .with_file(
            "/scripts/install_spark.sh",
            client.host().file("./dagger/spark-container/install_spark.sh"),
            permissions=755,
        )
        .with_exec(["./spark_setup/install_spark.sh"])
        .with_file("/usr/spark/conf/hive-site.xml", spark_dir.file("/hive-site.xml"))
        .with_file("/usr/spark/conf/spark-defaults.conf", spark_dir.file("spark-defaults.conf"))
    )

    # postgres is the metastore here
    pg_ctr, pg_host = get_postgres_container(client)

    spark_ctr = (
        spark_ctr_base.with_service_binding(alias=pg_host, service=pg_ctr)
        .with_exec(
            [
                "/scripts/entrypoint.sh",
                "--class",
                "org.apache.spark.sql.hive.thriftserver.HiveThriftServer2",
                "--name",
                "Thrift JDBC/ODBC Server",
            ]
        )
        .with_exposed_port(10000)
        .as_service()
    )

    return spark_ctr, "spark_db"


async def test_spark(test_args):
    async with dagger.Connection(dagger.Config(log_output=sys.stderr)) as client:

        # create cache volumes, these are persisted between runs saving time when developing locally
        tst_container = (
            client.container(platform=dagger.Platform("linux/amd64"))
            .from_("python:3.10-slim")
            .with_mounted_cache("/var/cache/apt/archives", client.cache_volume("os_reqs"))
            .with_mounted_cache("/root/.cache/pip", client.cache_volume("pip"))
        )

        # install system dependencies first so any local changes don't invalidate the cache
        tst_container = (
            tst_container.with_workdir("/")
            .with_directory("/scripts", client.host().directory("./dagger/scripts"))
            .with_exec(["./scripts/install_os_reqs.sh"])
            .with_exec(["pip", "install", "-U", "pip", "hatch"])
            .with_(env_variables(TESTING_ENV_VARS))
        )

        # copy project files into image
        tst_container = (
            tst_container.with_workdir("/")
            .with_directory("/src/src/dbt", client.host().directory("./src/dbt"))
            .with_directory("/src/tests", client.host().directory("./tests"))
            .with_directory(
                "/src",
                client.host().directory(
                    "./",
                    include=[
                        "pyproject.toml",
                        "hatch.toml",
                        "License.md",  # referenced in build metadata
                        "README.md",  # referenced in build metadata
                        "test.env",  # may not exist locally, does not exist in ci
                    ],
                ),
            )
        )

        # install profile-specific system dependencies last since tests usually rotate through profiles
        if test_args.profile == "apache_spark":
            spark_ctr, spark_host = get_spark_container(client)
            tst_container = tst_container.with_service_binding(alias=spark_host, service=spark_ctr)

        elif test_args.profile in [
            "databricks_cluster",
            "databricks_sql_endpoint",
            "spark_http_odbc",
        ]:
            tst_container = tst_container.with_workdir("/").with_exec(
                ["./scripts/configure_odbc.sh"]
            )

        elif test_args.profile == "spark_session":
            tst_container = tst_container.with_exec(["./scripts/install_jdk.sh"])

        # run the tests
        result = (
            await tst_container.with_workdir("/src")
            .with_exec(
                ["hatch", "run", "pytest", "--profile", test_args.profile, test_args.test_path]
            )
            .stdout()
        )

        return result


# TODO: update this to align more closely with the pytest api, e.g. --test-path should be an arg instead of a kwarg
parser = argparse.ArgumentParser()
parser.add_argument("--profile", required=True, type=str)
parser.add_argument("--test-path", required=False, type=str, default="tests/functional/adapter")
args = parser.parse_args()

anyio.run(test_spark, args)



from unittest import mock

import dbt.adapters.__about__
import dbt.adapters.spark.__version__

from dbt.adapters.spark.impl import SparkAdapter
from dbt.adapters.base.relation import AdapterTrackingRelationInfo


def assert_telemetry_data(adapter_type: str, file_format: str):
    table_formats_within_file_formats = ["delta", "iceberg", "hive", "hudi"]
    expected_table_format = None
    if file_format in table_formats_within_file_formats:
        expected_table_format = file_format

    mock_model_config = mock.MagicMock()
    mock_model_config._extra = mock.MagicMock()
    mock_model_config._extra = {
        "adapter_type": adapter_type,
        "file_format": file_format,
    }

    res = SparkAdapter.get_adapter_run_info(mock_model_config)

    assert res.adapter_name == adapter_type
    assert res.base_adapter_version == dbt.adapters.__about__.version
    assert res.adapter_version == dbt.adapters.spark.__version__.version

    assert res.model_adapter_details == {
        "adapter_type": adapter_type,
        "table_format": expected_table_format,
    }

    assert type(res) is AdapterTrackingRelationInfo


def test_telemetry_with_spark_details():
    spark_file_formats = [
        "text",
        "csv",
        "json",
        "jdbc",
        "parquet",
        "orc",
        "hive",
        "delta",
        "iceberg",
        "libsvm",
        "hudi",
    ]
    for file_format in spark_file_formats:
        assert_telemetry_data("spark", file_format)



from dbt.adapters.spark.connections import SparkConnectionMethod, SparkCredentials


def test_credentials_server_side_parameters_keys_and_values_are_strings() -> None:
    credentials = SparkCredentials(
        host="localhost",
        method=SparkConnectionMethod.THRIFT,  # type:ignore
        database="tests",
        schema="tests",
        server_side_parameters={"spark.configuration": "10"},
    )
    assert credentials.server_side_parameters["spark.configuration"] == "10"


def test_credentials_server_side_parameters_ansi_disabled_cannot_be_overridden() -> None:
    credentials = SparkCredentials(
        host="localhost",
        method=SparkConnectionMethod.THRIFT,  # type:ignore
        database="tests",
        schema="tests",
        server_side_parameters={"spark.sql.ansi.enabled": "true"},
    )
    assert credentials.server_side_parameters["spark.sql.ansi.enabled"] == "false"


def test_credentials_server_side_parameters_ansi_disabled_default() -> None:
    credentials = SparkCredentials(
        host="localhost",
        method=SparkConnectionMethod.THRIFT,  # type:ignore
        database="tests",
        schema="tests",
    )
    assert credentials.server_side_parameters["spark.sql.ansi.enabled"] == "false"



from base64 import standard_b64encode
from os import getenv
from time import sleep

import pytest

try:
    from pyhive import hive
    from thrift.transport.THttpClient import THttpClient
    from thrift.Thrift import TApplicationException
except ImportError:
    pass

from dbt.adapters.spark.connections import SparkConnectionManager


HOST = "https://" + getenv("DBT_DATABRICKS_HOST_NAME")
CLUSTER = getenv("DBT_DATABRICKS_CLUSTER_NAME")
TOKEN = getenv("DBT_DATABRICKS_TOKEN")
PORT = 443
ORGANIZATION = "0"


# Running this should prevent tests from needing to be retried because the Databricks cluster isn't available
@pytest.fixture(scope="session", autouse=True)
def start_databricks_cluster(request):

    profile = request.config.getoption("--profile")

    if profile.startswith("databricks"):
        _wait_for_databricks_cluster()

    yield


def _wait_for_databricks_cluster() -> None:
    """
    It takes roughly 3-5 minutes for the cluster to start, to be safe we'll wait for 10 minutes
    """
    transport_client = _transport_client()

    for _ in range(20):
        try:
            hive.connect(thrift_transport=transport_client)
            return
        except TApplicationException:
            sleep(30)

    raise Exception("Databricks cluster did not start in time")


def _transport_client():
    conn_url = SparkConnectionManager.SPARK_CONNECTION_URL.format(
        host=HOST,
        cluster=CLUSTER,
        port=PORT,
        organization=ORGANIZATION,
    )

    transport_client = THttpClient(conn_url)
    raw_token = f"token:{TOKEN}".encode()
    token = standard_b64encode(raw_token).decode()
    transport_client.setCustomHeaders({"Authorization": f"Basic {token}"})
    return transport_client



import unittest
import pytest
from multiprocessing import get_context
from unittest import mock

from dbt.exceptions import DbtRuntimeError
from agate import Row
from pyhive import hive
from dbt.adapters.spark import SparkAdapter, SparkRelation
from .utils import config_from_parts_or_dicts

ENFORCED_SPARK_CONFIG = {"spark.sql.ansi.enabled": "false"}


class TestSparkAdapter(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def set_up_fixtures(
        self,
        target_http,
        target_odbc_with_extra_conn,
        target_thrift,
        target_thrift_kerberos,
        target_odbc_sql_endpoint,
        target_odbc_cluster,
        target_use_ssl_thrift,
        base_project_cfg,
    ):
        self.base_project_cfg = base_project_cfg
        self.target_http = target_http
        self.target_odbc_with_extra_conn = target_odbc_with_extra_conn
        self.target_odbc_sql_endpoint = target_odbc_sql_endpoint
        self.target_odbc_cluster = target_odbc_cluster
        self.target_thrift = target_thrift
        self.target_thrift_kerberos = target_thrift_kerberos
        self.target_use_ssl_thrift = target_use_ssl_thrift

    def test_http_connection(self):
        adapter = SparkAdapter(self.target_http, get_context("spawn"))

        def hive_http_connect(thrift_transport, configuration):
            self.assertEqual(thrift_transport.scheme, "https")
            self.assertEqual(thrift_transport.port, 443)
            self.assertEqual(thrift_transport.host, "myorg.sparkhost.com")
            self.assertEqual(
                thrift_transport.path, "/sql/protocolv1/o/0123456789/01234-23423-coffeetime"
            )
            self.assertEqual(configuration["spark.driver.memory"], "4g")

        # with mock.patch.object(hive, 'connect', new=hive_http_connect):
        with mock.patch("dbt.adapters.spark.connections.hive.connect", new=hive_http_connect):
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertEqual(connection.credentials.cluster, "01234-23423-coffeetime")
            self.assertEqual(connection.credentials.token, "abc123")
            self.assertEqual(connection.credentials.schema, "analytics")
            self.assertIsNone(connection.credentials.database)

    def test_thrift_connection(self):
        config = self.target_thrift
        adapter = SparkAdapter(config, get_context("spawn"))

        def hive_thrift_connect(
            host, port, username, auth, kerberos_service_name, password, configuration
        ):
            self.assertEqual(host, "myorg.sparkhost.com")
            self.assertEqual(port, 10001)
            self.assertEqual(username, "dbt")
            self.assertIsNone(auth)
            self.assertIsNone(kerberos_service_name)
            self.assertIsNone(password)
            self.assertDictEqual(configuration, ENFORCED_SPARK_CONFIG)

        with mock.patch.object(hive, "connect", new=hive_thrift_connect):
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertEqual(connection.credentials.schema, "analytics")
            self.assertIsNone(connection.credentials.database)

    def test_thrift_ssl_connection(self):
        adapter = SparkAdapter(self.target_use_ssl_thrift, get_context("spawn"))

        def hive_thrift_connect(thrift_transport, configuration):
            self.assertIsNotNone(thrift_transport)
            transport = thrift_transport._trans
            self.assertEqual(transport.host, "myorg.sparkhost.com")
            self.assertEqual(transport.port, 10001)
            self.assertDictEqual(configuration, ENFORCED_SPARK_CONFIG)

        with mock.patch.object(hive, "connect", new=hive_thrift_connect):
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertEqual(connection.credentials.schema, "analytics")
            self.assertIsNone(connection.credentials.database)

    def test_thrift_connection_kerberos(self):
        adapter = SparkAdapter(self.target_thrift_kerberos, get_context("spawn"))

        def hive_thrift_connect(
            host, port, username, auth, kerberos_service_name, password, configuration
        ):
            self.assertEqual(host, "myorg.sparkhost.com")
            self.assertEqual(port, 10001)
            self.assertEqual(username, "dbt")
            self.assertEqual(auth, "KERBEROS")
            self.assertEqual(kerberos_service_name, "hive")
            self.assertIsNone(password)
            self.assertDictEqual(configuration, ENFORCED_SPARK_CONFIG)

        with mock.patch.object(hive, "connect", new=hive_thrift_connect):
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertEqual(connection.credentials.schema, "analytics")
            self.assertIsNone(connection.credentials.database)

    def test_odbc_cluster_connection(self):
        adapter = SparkAdapter(self.target_odbc_cluster, get_context("spawn"))

        def pyodbc_connect(connection_str, autocommit):
            self.assertTrue(autocommit)
            self.assertIn("driver=simba;", connection_str.lower())
            self.assertIn("port=443;", connection_str.lower())
            self.assertIn("host=myorg.sparkhost.com;", connection_str.lower())
            self.assertIn(
                "httppath=/sql/protocolv1/o/0123456789/01234-23423-coffeetime;",
                connection_str.lower(),
            )  # noqa

        with mock.patch(
            "dbt.adapters.spark.connections.pyodbc.connect", new=pyodbc_connect
        ):  # noqa
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertEqual(connection.credentials.cluster, "01234-23423-coffeetime")
            self.assertEqual(connection.credentials.token, "abc123")
            self.assertEqual(connection.credentials.schema, "analytics")
            self.assertIsNone(connection.credentials.database)

    def test_odbc_endpoint_connection(self):
        adapter = SparkAdapter(self.target_odbc_sql_endpoint, get_context("spawn"))

        def pyodbc_connect(connection_str, autocommit):
            self.assertTrue(autocommit)
            self.assertIn("driver=simba;", connection_str.lower())
            self.assertIn("port=443;", connection_str.lower())
            self.assertIn("host=myorg.sparkhost.com;", connection_str.lower())
            self.assertIn(
                "httppath=/sql/1.0/endpoints/012342342393920a;", connection_str.lower()
            )  # noqa

        with mock.patch(
            "dbt.adapters.spark.connections.pyodbc.connect", new=pyodbc_connect
        ):  # noqa
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertEqual(connection.credentials.endpoint, "012342342393920a")
            self.assertEqual(connection.credentials.token, "abc123")
            self.assertEqual(connection.credentials.schema, "analytics")
            self.assertIsNone(connection.credentials.database)

    def test_odbc_with_extra_connection_string(self):
        adapter = SparkAdapter(self.target_odbc_with_extra_conn, get_context("spawn"))

        def pyodbc_connect(connection_str, autocommit):
            self.assertTrue(autocommit)
            self.assertIn("driver=simba;", connection_str.lower())
            self.assertIn("port=443;", connection_str.lower())
            self.assertIn("host=myorg.sparkhost.com;", connection_str.lower())
            self.assertIn("someExtraValues", connection_str)

        with mock.patch(
            "dbt.adapters.spark.connections.pyodbc.connect", new=pyodbc_connect
        ):  # noqa
            connection = adapter.acquire_connection("dummy")
            connection.handle  # trigger lazy-load

            self.assertEqual(connection.state, "open")
            self.assertIsNotNone(connection.handle)
            self.assertIsNone(connection.credentials.database)

    def test_parse_relation(self):
        self.maxDiff = None
        rel_type = SparkRelation.get_relation_type.Table

        relation = SparkRelation.create(
            schema="default_schema", identifier="mytable", type=rel_type
        )
        assert relation.database is None

        # Mimics the output of Spark with a DESCRIBE TABLE EXTENDED
        plain_rows = [
            ("col1", "decimal(22,0)"),
            (
                "col2",
                "string",
            ),
            ("dt", "date"),
            ("struct_col", "struct<struct_inner_col:string>"),
            ("# Partition Information", "data_type"),
            ("# col_name", "data_type"),
            ("dt", "date"),
            (None, None),
            ("# Detailed Table Information", None),
            ("Database", None),
            ("Owner", "root"),
            ("Created Time", "Wed Feb 04 18:15:00 UTC 1815"),
            ("Last Access", "Wed May 20 19:25:00 UTC 1925"),
            ("Type", "MANAGED"),
            ("Provider", "delta"),
            ("Location", "/mnt/vo"),
            ("Serde Library", "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe"),
            ("InputFormat", "org.apache.hadoop.mapred.SequenceFileInputFormat"),
            ("OutputFormat", "org.apache.hadoop.hive.ql.io.HiveSequenceFileOutputFormat"),
            ("Partition Provider", "Catalog"),
        ]

        input_cols = [Row(keys=["col_name", "data_type"], values=r) for r in plain_rows]

        rows = SparkAdapter(self.target_http, get_context("spawn")).parse_describe_extended(
            relation, input_cols
        )
        self.assertEqual(len(rows), 4)
        self.assertEqual(
            rows[0].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "col1",
                "column_index": 0,
                "dtype": "decimal(22,0)",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
            },
        )

        self.assertEqual(
            rows[1].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "col2",
                "column_index": 1,
                "dtype": "string",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
            },
        )

        self.assertEqual(
            rows[2].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "dt",
                "column_index": 2,
                "dtype": "date",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
            },
        )

        self.assertEqual(
            rows[3].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "struct_col",
                "column_index": 3,
                "dtype": "struct<struct_inner_col:string>",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
            },
        )

    def test_parse_relation_with_integer_owner(self):
        self.maxDiff = None
        rel_type = SparkRelation.get_relation_type.Table

        relation = SparkRelation.create(
            schema="default_schema", identifier="mytable", type=rel_type
        )
        assert relation.database is None

        # Mimics the output of Spark with a DESCRIBE TABLE EXTENDED
        plain_rows = [
            ("col1", "decimal(22,0)"),
            ("# Detailed Table Information", None),
            ("Owner", 1234),
        ]

        input_cols = [Row(keys=["col_name", "data_type"], values=r) for r in plain_rows]

        rows = SparkAdapter(self.target_http, get_context("spawn")).parse_describe_extended(
            relation, input_cols
        )

        self.assertEqual(rows[0].to_column_dict().get("table_owner"), "1234")

    def test_parse_relation_with_statistics(self):
        self.maxDiff = None
        rel_type = SparkRelation.get_relation_type.Table

        relation = SparkRelation.create(
            schema="default_schema", identifier="mytable", type=rel_type
        )
        assert relation.database is None

        # Mimics the output of Spark with a DESCRIBE TABLE EXTENDED
        plain_rows = [
            ("col1", "decimal(22,0)"),
            ("# Partition Information", "data_type"),
            (None, None),
            ("# Detailed Table Information", None),
            ("Database", None),
            ("Owner", "root"),
            ("Created Time", "Wed Feb 04 18:15:00 UTC 1815"),
            ("Last Access", "Wed May 20 19:25:00 UTC 1925"),
            ("Statistics", "1109049927 bytes, 14093476 rows"),
            ("Type", "MANAGED"),
            ("Provider", "delta"),
            ("Location", "/mnt/vo"),
            ("Serde Library", "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe"),
            ("InputFormat", "org.apache.hadoop.mapred.SequenceFileInputFormat"),
            ("OutputFormat", "org.apache.hadoop.hive.ql.io.HiveSequenceFileOutputFormat"),
            ("Partition Provider", "Catalog"),
        ]

        input_cols = [Row(keys=["col_name", "data_type"], values=r) for r in plain_rows]

        rows = SparkAdapter(self.target_http, get_context("spawn")).parse_describe_extended(
            relation, input_cols
        )
        self.assertEqual(len(rows), 1)
        self.assertEqual(
            rows[0].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "col1",
                "column_index": 0,
                "dtype": "decimal(22,0)",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 1109049927,
                "stats:rows:description": "",
                "stats:rows:include": True,
                "stats:rows:label": "rows",
                "stats:rows:value": 14093476,
            },
        )

    def test_relation_with_database(self):
        adapter = SparkAdapter(self.target_http, get_context("spawn"))
        # fine
        adapter.Relation.create(schema="different", identifier="table")
        with self.assertRaises(DbtRuntimeError):
            # not fine - database set
            adapter.Relation.create(database="something", schema="different", identifier="table")

    def test_profile_with_database(self):
        profile = {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "http",
                    # not allowed
                    "database": "analytics2",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 443,
                    "token": "abc123",
                    "organization": "0123456789",
                    "cluster": "01234-23423-coffeetime",
                }
            },
            "target": "test",
        }
        with self.assertRaises(DbtRuntimeError):
            config_from_parts_or_dicts(self.base_project_cfg, profile)

    def test_profile_with_cluster_and_sql_endpoint(self):
        profile = {
            "outputs": {
                "test": {
                    "type": "spark",
                    "method": "odbc",
                    "schema": "analytics",
                    "host": "myorg.sparkhost.com",
                    "port": 443,
                    "token": "abc123",
                    "organization": "0123456789",
                    "cluster": "01234-23423-coffeetime",
                    "endpoint": "0123412341234e",
                }
            },
            "target": "test",
        }
        with self.assertRaises(DbtRuntimeError):
            config_from_parts_or_dicts(self.base_project_cfg, profile)

    def test_parse_columns_from_information_with_table_type_and_delta_provider(self):
        self.maxDiff = None
        rel_type = SparkRelation.get_relation_type.Table

        # Mimics the output of Spark in the information column
        information = (
            "Database: default_schema\n"
            "Table: mytable\n"
            "Owner: root\n"
            "Created Time: Wed Feb 04 18:15:00 UTC 1815\n"
            "Last Access: Wed May 20 19:25:00 UTC 1925\n"
            "Created By: Spark 3.0.1\n"
            "Type: MANAGED\n"
            "Provider: delta\n"
            "Statistics: 123456789 bytes\n"
            "Location: /mnt/vo\n"
            "Serde Library: org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe\n"
            "InputFormat: org.apache.hadoop.mapred.SequenceFileInputFormat\n"
            "OutputFormat: org.apache.hadoop.hive.ql.io.HiveSequenceFileOutputFormat\n"
            "Partition Provider: Catalog\n"
            "Partition Columns: [`dt`]\n"
            "Schema: root\n"
            " |-- col1: decimal(22,0) (nullable = true)\n"
            " |-- col2: string (nullable = true)\n"
            " |-- dt: date (nullable = true)\n"
            " |-- struct_col: struct (nullable = true)\n"
            " |    |-- struct_inner_col: string (nullable = true)\n"
        )
        relation = SparkRelation.create(
            schema="default_schema", identifier="mytable", type=rel_type, information=information
        )

        columns = SparkAdapter(
            self.target_http, get_context("spawn")
        ).parse_columns_from_information(relation)
        self.assertEqual(len(columns), 4)
        self.assertEqual(
            columns[0].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "col1",
                "column_index": 0,
                "dtype": "decimal(22,0)",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 123456789,
            },
        )

        self.assertEqual(
            columns[3].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "struct_col",
                "column_index": 3,
                "dtype": "struct",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 123456789,
            },
        )

    def test_parse_columns_from_information_with_view_type(self):
        self.maxDiff = None
        rel_type = SparkRelation.get_relation_type.View
        information = (
            "Database: default_schema\n"
            "Table: myview\n"
            "Owner: root\n"
            "Created Time: Wed Feb 04 18:15:00 UTC 1815\n"
            "Last Access: UNKNOWN\n"
            "Created By: Spark 3.0.1\n"
            "Type: VIEW\n"
            "View Text: WITH base (\n"
            "    SELECT * FROM source_table\n"
            ")\n"
            "SELECT col1, col2, dt FROM base\n"
            "View Original Text: WITH base (\n"
            "    SELECT * FROM source_table\n"
            ")\n"
            "SELECT col1, col2, dt FROM base\n"
            "View Catalog and Namespace: spark_catalog.default\n"
            "View Query Output Columns: [col1, col2, dt]\n"
            "Table Properties: [view.query.out.col.1=col1, view.query.out.col.2=col2, "
            "transient_lastDdlTime=1618324324, view.query.out.col.3=dt, "
            "view.catalogAndNamespace.part.0=spark_catalog, "
            "view.catalogAndNamespace.part.1=default]\n"
            "Serde Library: org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe\n"
            "InputFormat: org.apache.hadoop.mapred.SequenceFileInputFormat\n"
            "OutputFormat: org.apache.hadoop.hive.ql.io.HiveSequenceFileOutputFormat\n"
            "Storage Properties: [serialization.format=1]\n"
            "Schema: root\n"
            " |-- col1: decimal(22,0) (nullable = true)\n"
            " |-- col2: string (nullable = true)\n"
            " |-- dt: date (nullable = true)\n"
            " |-- struct_col: struct (nullable = true)\n"
            " |    |-- struct_inner_col: string (nullable = true)\n"
        )
        relation = SparkRelation.create(
            schema="default_schema", identifier="myview", type=rel_type, information=information
        )

        columns = SparkAdapter(
            self.target_http, get_context("spawn")
        ).parse_columns_from_information(relation)
        self.assertEqual(len(columns), 4)
        self.assertEqual(
            columns[1].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "col2",
                "column_index": 1,
                "dtype": "string",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
            },
        )

        self.assertEqual(
            columns[3].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "struct_col",
                "column_index": 3,
                "dtype": "struct",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
            },
        )

    def test_parse_columns_from_information_with_table_type_and_parquet_provider(self):
        self.maxDiff = None
        rel_type = SparkRelation.get_relation_type.Table

        information = (
            "Database: default_schema\n"
            "Table: mytable\n"
            "Owner: root\n"
            "Created Time: Wed Feb 04 18:15:00 UTC 1815\n"
            "Last Access: Wed May 20 19:25:00 UTC 1925\n"
            "Created By: Spark 3.0.1\n"
            "Type: MANAGED\n"
            "Provider: parquet\n"
            "Statistics: 1234567890 bytes, 12345678 rows\n"
            "Location: /mnt/vo\n"
            "Serde Library: org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe\n"
            "InputFormat: org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat\n"
            "OutputFormat: org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat\n"
            "Schema: root\n"
            " |-- col1: decimal(22,0) (nullable = true)\n"
            " |-- col2: string (nullable = true)\n"
            " |-- dt: date (nullable = true)\n"
            " |-- struct_col: struct (nullable = true)\n"
            " |    |-- struct_inner_col: string (nullable = true)\n"
        )
        relation = SparkRelation.create(
            schema="default_schema", identifier="mytable", type=rel_type, information=information
        )

        columns = SparkAdapter(
            self.target_http, get_context("spawn")
        ).parse_columns_from_information(relation)
        self.assertEqual(len(columns), 4)

        self.assertEqual(
            columns[2].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "dt",
                "column_index": 2,
                "dtype": "date",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 1234567890,
                "stats:rows:description": "",
                "stats:rows:include": True,
                "stats:rows:label": "rows",
                "stats:rows:value": 12345678,
            },
        )

        self.assertEqual(
            columns[3].to_column_dict(omit_none=False),
            {
                "table_database": None,
                "table_schema": relation.schema,
                "table_name": relation.name,
                "table_type": rel_type,
                "table_owner": "root",
                "column": "struct_col",
                "column_index": 3,
                "dtype": "struct",
                "numeric_scale": None,
                "numeric_precision": None,
                "char_size": None,
                "stats:bytes:description": "",
                "stats:bytes:include": True,
                "stats:bytes:label": "bytes",
                "stats:bytes:value": 1234567890,
                "stats:rows:description": "",
                "stats:rows:include": True,
                "stats:rows:label": "rows",
                "stats:rows:value": 12345678,
            },
        )



from dbt.tests.adapter.empty.test_empty import BaseTestEmpty, BaseTestEmptyInlineSourceRef


class TestSparkEmpty(BaseTestEmpty):
    pass


class TestSparkEmptyInlineSourceRef(BaseTestEmptyInlineSourceRef):
    pass



import pytest

from dbt.tests.adapter.store_test_failures_tests import basic
from dbt.tests.adapter.store_test_failures_tests.test_store_test_failures import (
    StoreTestFailuresBase,
    TEST_AUDIT_SCHEMA_SUFFIX,
)


@pytest.mark.skip_profile(
    "spark_session", "databricks_cluster", "databricks_sql_endpoint", "spark_http_odbc"
)
class TestSparkStoreTestFailures(StoreTestFailuresBase):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seeds": {
                "quote_columns": True,
            },
            "tests": {"+schema": TEST_AUDIT_SCHEMA_SUFFIX, "+store_failures": True},
        }

    @pytest.fixture(scope="function", autouse=True)
    def teardown_method(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database,
                schema=f"{project.test_schema}_{TEST_AUDIT_SCHEMA_SUFFIX}",
            )

            project.adapter.drop_schema(relation)

    def test_store_and_assert(self, project):
        self.run_tests_store_one_failure(project)
        self.run_tests_store_failures_and_assert(project)


@pytest.mark.skip_profile(
    "apache_spark", "spark_session", "databricks_sql_endpoint", "spark_http_odbc"
)
class TestSparkStoreTestFailuresWithDelta(StoreTestFailuresBase):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seeds": {
                "quote_columns": False,
                "test": self.column_type_overrides(),
                "+file_format": "delta",
            },
            "tests": {
                "+schema": TEST_AUDIT_SCHEMA_SUFFIX,
                "+store_failures": True,
                "+file_format": "delta",
            },
        }

    @pytest.fixture(scope="function", autouse=True)
    def teardown_method(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database,
                schema=f"{project.test_schema}_{TEST_AUDIT_SCHEMA_SUFFIX}",
            )

            project.adapter.drop_schema(relation)

    def test_store_and_assert_failure_with_delta(self, project):
        self.run_tests_store_one_failure(project)
        self.run_tests_store_failures_and_assert(project)


@pytest.mark.skip_profile("spark_session")
class TestStoreTestFailuresAsInteractions(basic.StoreTestFailuresAsInteractions):
    pass


@pytest.mark.skip_profile("spark_session")
class TestStoreTestFailuresAsProjectLevelOff(basic.StoreTestFailuresAsProjectLevelOff):
    pass


@pytest.mark.skip_profile("spark_session")
class TestStoreTestFailuresAsProjectLevelView(basic.StoreTestFailuresAsProjectLevelView):
    pass


@pytest.mark.skip_profile("spark_session")
class TestStoreTestFailuresAsGeneric(basic.StoreTestFailuresAsGeneric):
    pass


@pytest.mark.skip_profile("spark_session")
class TestStoreTestFailuresAsProjectLevelEphemeral(basic.StoreTestFailuresAsProjectLevelEphemeral):
    pass


@pytest.mark.skip_profile("spark_session")
class TestStoreTestFailuresAsExceptions(basic.StoreTestFailuresAsExceptions):
    pass



import os

import pytest


@pytest.fixture(scope="session")
def dbt_profile_target(request):

    creds = {
        "apache_spark": apache_spark_target,
        "spark_http_odbc": spark_http_odbc_target,
        "spark_session": spark_session_target,
        "databricks_cluster": databricks_cluster_target,
        "databricks_http_cluster": databricks_http_cluster_target,
        "databricks_sql_endpoint": databricks_sql_endpoint_target,
    }

    profile = request.config.getoption("--profile")

    try:
        return creds[profile]()
    except KeyError:
        raise ValueError(f"Invalid profile type '{profile}'")


@pytest.fixture(autouse=True)
def skip_by_profile_type(request):
    profile_type = request.config.getoption("--profile")
    if request.node.get_closest_marker("skip_profile"):
        for skip_profile_type in request.node.get_closest_marker("skip_profile").args:
            if skip_profile_type == profile_type:
                pytest.skip(f"skipped on '{profile_type}' profile")


def apache_spark_target():
    return {
        "type": "spark",
        "host": "spark_db",
        "user": "dbt",
        "method": "thrift",
        "port": 10000,
        "connect_retries": 2,
        "connect_timeout": 3,
        "retry_all": False,
    }


def databricks_cluster_target():
    return {
        "type": "spark",
        "method": "odbc",
        "host": os.getenv("DBT_DATABRICKS_HOST_NAME"),
        "cluster": os.getenv("DBT_DATABRICKS_CLUSTER_NAME"),
        "token": os.getenv("DBT_DATABRICKS_TOKEN"),
        "driver": os.getenv("ODBC_DRIVER"),
        "port": 443,
        "connect_retries": 3,
        "connect_timeout": 5,
        "retry_all": False,
        "user": os.getenv("DBT_DATABRICKS_USER"),
    }


def databricks_sql_endpoint_target():
    return {
        "type": "spark",
        "method": "odbc",
        "host": os.getenv("DBT_DATABRICKS_HOST_NAME"),
        "endpoint": os.getenv("DBT_DATABRICKS_ENDPOINT"),
        "token": os.getenv("DBT_DATABRICKS_TOKEN"),
        "driver": os.getenv("ODBC_DRIVER"),
        "port": 443,
        "connect_retries": 3,
        "connect_timeout": 5,
        "retry_all": True,
    }


def databricks_http_cluster_target():
    return {
        "type": "spark",
        "host": os.getenv("DBT_DATABRICKS_HOST_NAME"),
        "cluster": os.getenv("DBT_DATABRICKS_CLUSTER_NAME"),
        "token": os.getenv("DBT_DATABRICKS_TOKEN"),
        "method": "http",
        "port": 443,
        "connect_retries": 3,
        "connect_timeout": 5,
        "retry_all": False,
        "user": os.getenv("DBT_DATABRICKS_USER"),
    }


def spark_session_target():
    return {
        "type": "spark",
        "host": "localhost",
        "method": "session",
    }


def spark_http_odbc_target():
    return {
        "type": "spark",
        "method": "odbc",
        "host": os.getenv("DBT_DATABRICKS_HOST_NAME"),
        "port": 443,
        "driver": os.getenv("ODBC_DRIVER"),
        "connection_string_suffix": f'UID=token;PWD={os.getenv("DBT_DATABRICKS_TOKEN")};HTTPPath=/sql/1.0/endpoints/{os.getenv("DBT_DATABRICKS_ENDPOINT")};AuthMech=3;SparkServerType=3',
        "connect_retries": 3,
        "connect_timeout": 5,
        "retry_all": True,
    }



from tests.functional.fixtures import (
    dbt_profile_target,
    skip_by_profile_type,
    start_databricks_cluster,
)


pytest_plugins = ["dbt.tests.fixtures.project"]


def pytest_addoption(parser):
    parser.addoption("--profile", action="store", default="apache_spark", type=str)


# Using @pytest.mark.skip_profile('apache_spark') uses the 'skip_by_profile_type'
# autouse fixture below
def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "skip_profile(profile): skip test for the given profile",
    )



import pytest
from dbt.tests.adapter.constraints.test_constraints import (
    BaseModelConstraintsRuntimeEnforcement,
    BaseTableConstraintsColumnsEqual,
    BaseViewConstraintsColumnsEqual,
    BaseIncrementalConstraintsColumnsEqual,
    BaseConstraintsRuntimeDdlEnforcement,
    BaseConstraintsRollback,
    BaseIncrementalConstraintsRuntimeDdlEnforcement,
    BaseIncrementalConstraintsRollback,
    BaseConstraintQuotedColumn,
)
from dbt.tests.adapter.constraints.fixtures import (
    constrained_model_schema_yml,
    my_model_sql,
    my_model_wrong_order_sql,
    my_model_wrong_name_sql,
    model_schema_yml,
    my_model_view_wrong_order_sql,
    my_model_view_wrong_name_sql,
    my_model_incremental_wrong_order_sql,
    my_model_incremental_wrong_name_sql,
    my_incremental_model_sql,
    model_fk_constraint_schema_yml,
    my_model_wrong_order_depends_on_fk_sql,
    foreign_key_model_sql,
    my_model_incremental_wrong_order_depends_on_fk_sql,
    my_model_with_quoted_column_name_sql,
    model_quoted_column_schema_yml,
)

# constraints are enforced via 'alter' statements that run after table creation
_expected_sql_spark = """
create or replace table <model_identifier>
    using delta
    as
select
  id,
  color,
  date_day
from

(
    -- depends_on: <foreign_key_model_identifier>
    select
    'blue' as color,
    1 as id,
    '2019-01-01' as date_day ) as model_subq
"""

_expected_sql_spark_model_constraints = """
create or replace table <model_identifier>
    using delta
    as
select
  id,
  color,
  date_day
from

(
    -- depends_on: <foreign_key_model_identifier>
    select
    'blue' as color,
    1 as id,
    '2019-01-01' as date_day ) as model_subq
"""

# Different on Spark:
# - does not support a data type named 'text' (TODO handle this in the base test classes using string_type
constraints_yml = model_schema_yml.replace("text", "string").replace("primary key", "")
model_fk_constraint_schema_yml = model_fk_constraint_schema_yml.replace("text", "string").replace(
    "primary key", ""
)
model_constraints_yml = constrained_model_schema_yml.replace("text", "string")


class PyodbcSetup:
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
            }
        }

    @pytest.fixture
    def string_type(self):
        return "STR"

    @pytest.fixture
    def int_type(self):
        return "INT"

    @pytest.fixture
    def schema_string_type(self):
        return "STRING"

    @pytest.fixture
    def schema_int_type(self):
        return "INT"

    @pytest.fixture
    def data_types(self, int_type, schema_int_type, string_type, schema_string_type):
        # sql_column_value, schema_data_type, error_data_type
        return [
            ["1", schema_int_type, int_type],
            ['"1"', schema_string_type, string_type],
            ["true", "boolean", "BOOL"],
            ['array("1","2","3")', "string", string_type],
            ["array(1,2,3)", "string", string_type],
            ["6.45", "decimal", "DECIMAL"],
            ["cast('2019-01-01' as date)", "date", "DATE"],
            ["cast('2019-01-01' as timestamp)", "timestamp", "DATETIME"],
        ]


class DatabricksHTTPSetup:
    @pytest.fixture
    def string_type(self):
        return "STRING_TYPE"

    @pytest.fixture
    def int_type(self):
        return "INT_TYPE"

    @pytest.fixture
    def schema_string_type(self):
        return "STRING"

    @pytest.fixture
    def schema_int_type(self):
        return "INT"

    @pytest.fixture
    def data_types(self, int_type, schema_int_type, string_type, schema_string_type):
        # sql_column_value, schema_data_type, error_data_type
        return [
            ["1", schema_int_type, int_type],
            ['"1"', schema_string_type, string_type],
            ["true", "boolean", "BOOLEAN_TYPE"],
            ['array("1","2","3")', "array<string>", "ARRAY_TYPE"],
            ["array(1,2,3)", "array<int>", "ARRAY_TYPE"],
            ["cast('2019-01-01' as date)", "date", "DATE_TYPE"],
            ["cast('2019-01-01' as timestamp)", "timestamp", "TIMESTAMP_TYPE"],
            ["cast(1.0 AS DECIMAL(4, 2))", "decimal", "DECIMAL_TYPE"],
        ]


@pytest.mark.skip_profile("spark_session", "apache_spark", "databricks_http_cluster")
class TestSparkTableConstraintsColumnsEqualPyodbc(PyodbcSetup, BaseTableConstraintsColumnsEqual):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


@pytest.mark.skip_profile("spark_session", "apache_spark", "databricks_http_cluster")
class TestSparkViewConstraintsColumnsEqualPyodbc(PyodbcSetup, BaseViewConstraintsColumnsEqual):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_view_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_view_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


@pytest.mark.skip_profile("spark_session", "apache_spark", "databricks_http_cluster")
class TestSparkIncrementalConstraintsColumnsEqualPyodbc(
    PyodbcSetup, BaseIncrementalConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_incremental_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_incremental_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


@pytest.mark.skip_profile(
    "spark_session",
    "apache_spark",
    "databricks_sql_endpoint",
    "databricks_cluster",
    "spark_http_odbc",
)
class TestSparkTableConstraintsColumnsEqualDatabricksHTTP(
    DatabricksHTTPSetup, BaseTableConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


@pytest.mark.skip_profile(
    "spark_session",
    "apache_spark",
    "databricks_sql_endpoint",
    "databricks_cluster",
    "spark_http_odbc",
)
class TestSparkViewConstraintsColumnsEqualDatabricksHTTP(
    DatabricksHTTPSetup, BaseViewConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_view_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_view_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


@pytest.mark.skip_profile(
    "spark_session",
    "apache_spark",
    "databricks_sql_endpoint",
    "databricks_cluster",
    "spark_http_odbc",
)
class TestSparkIncrementalConstraintsColumnsEqualDatabricksHTTP(
    DatabricksHTTPSetup, BaseIncrementalConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_incremental_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_incremental_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


class BaseSparkConstraintsDdlEnforcementSetup:
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
            }
        }

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return _expected_sql_spark


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestSparkTableConstraintsDdlEnforcement(
    BaseSparkConstraintsDdlEnforcementSetup, BaseConstraintsRuntimeDdlEnforcement
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_wrong_order_depends_on_fk_sql,
            "foreign_key_model.sql": foreign_key_model_sql,
            "constraints_schema.yml": model_fk_constraint_schema_yml,
        }


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestSparkIncrementalConstraintsDdlEnforcement(
    BaseSparkConstraintsDdlEnforcementSetup, BaseIncrementalConstraintsRuntimeDdlEnforcement
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_incremental_wrong_order_depends_on_fk_sql,
            "foreign_key_model.sql": foreign_key_model_sql,
            "constraints_schema.yml": model_fk_constraint_schema_yml,
        }


@pytest.mark.skip_profile("spark_session", "apache_spark", "databricks_http_cluster")
class TestSparkConstraintQuotedColumn(PyodbcSetup, BaseConstraintQuotedColumn):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_with_quoted_column_name_sql,
            "constraints_schema.yml": model_quoted_column_schema_yml.replace(
                "text", "string"
            ).replace('"from"', "`from`"),
        }

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """
create or replace table <model_identifier>
    using delta
    as
select
  id,
  `from`,
  date_day
from

(
    select
    'blue' as `from`,
    1 as id,
    '2019-01-01' as date_day ) as model_subq
"""


class BaseSparkConstraintsRollbackSetup:
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
            }
        }

    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return [
            "violate the new CHECK constraint",
            "DELTA_NEW_CHECK_CONSTRAINT_VIOLATION",
            "DELTA_NEW_NOT_NULL_VIOLATION",
            "violate the new NOT NULL constraint",
            "(id > 0) violated by row with values:",  # incremental mats
            "DELTA_VIOLATE_CONSTRAINT_WITH_VALUES",  # incremental mats
            "NOT NULL constraint violated for col",
        ]

    def assert_expected_error_messages(self, error_message, expected_error_messages):
        # This needs to be ANY instead of ALL
        # The CHECK constraint is added before the NOT NULL constraint
        # and different connection types display/truncate the error message in different ways...
        assert any(msg in error_message for msg in expected_error_messages)


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestSparkTableConstraintsRollback(
    BaseSparkConstraintsRollbackSetup, BaseConstraintsRollback
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_sql,
            "constraints_schema.yml": constraints_yml,
        }

    # On Spark/Databricks, constraints are applied *after* the table is replaced.
    # We don't have any way to "rollback" the table to its previous happy state.
    # So the 'color' column will be updated to 'red', instead of 'blue'.
    @pytest.fixture(scope="class")
    def expected_color(self):
        return "red"


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestSparkIncrementalConstraintsRollback(
    BaseSparkConstraintsRollbackSetup, BaseIncrementalConstraintsRollback
):
    # color stays blue for incremental models since it's a new row that just
    # doesn't get inserted
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_incremental_model_sql,
            "constraints_schema.yml": constraints_yml,
        }


# TODO: Like the tests above, this does test that model-level constraints don't
# result in errors, but it does not verify that they are actually present in
# Spark and that the ALTER TABLE statement actually ran.
@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestSparkModelConstraintsRuntimeEnforcement(BaseModelConstraintsRuntimeEnforcement):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
            }
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_wrong_order_depends_on_fk_sql,
            "foreign_key_model.sql": foreign_key_model_sql,
            "constraints_schema.yml": model_fk_constraint_schema_yml,
        }

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return _expected_sql_spark_model_constraints



expected_append_csv = """
id,msg
1,hello
2,goodbye
2,yo
3,anyway
""".lstrip()

expected_overwrite_csv = """
id,msg
2,yo
3,anyway
""".lstrip()

expected_partial_upsert_csv = """
id,msg,color
1,hello,blue
2,yo,red
3,anyway,purple
""".lstrip()

expected_upsert_csv = """
id,msg
1,hello
2,yo
3,anyway
""".lstrip()



from contextlib import contextmanager

from dbt.adapters.contracts.connection import (
    AdapterResponse,
    ConnectionState,
    Connection,
    Credentials,
)
from dbt.adapters.events.logging import AdapterLogger
from dbt.adapters.exceptions import FailedToConnectError
from dbt.adapters.sql import SQLConnectionManager
from dbt_common.exceptions import DbtConfigError, DbtRuntimeError, DbtDatabaseError

from dbt_common.utils.encoding import DECIMALS
from dbt.adapters.spark import __version__

try:
    from TCLIService.ttypes import TOperationState as ThriftState
    from thrift.transport import THttpClient
    from pyhive import hive
except ImportError:
    ThriftState = None
    THttpClient = None
    hive = None
try:
    import pyodbc
except ImportError:
    pyodbc = None
from datetime import datetime
import sqlparams
from dbt_common.dataclass_schema import StrEnum
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Union, Tuple, List, Generator, Iterable, Sequence

from abc import ABC, abstractmethod

try:
    from thrift.transport.TSSLSocket import TSSLSocket
    from thrift.transport.TTransport import TTransportException
    import thrift
    import ssl
    import thrift_sasl
    from puresasl.client import SASLClient
except ImportError:
    TTransportException = None
    pass  # done deliberately: setting modules to None explicitly violates MyPy contracts by degrading type semantics

import base64
import time

logger = AdapterLogger("Spark")

NUMBERS = DECIMALS + (int, float)

# Exception types that indicate a lost connection during query polling
# Built once at module load time to avoid reconstructing on every poll
CONNECTION_LOST_EXCEPTIONS: Tuple[type, ...]
if TTransportException is not None:
    CONNECTION_LOST_EXCEPTIONS = (
        ConnectionResetError,
        BrokenPipeError,
        EOFError,
        TTransportException,
    )
else:
    CONNECTION_LOST_EXCEPTIONS = (
        ConnectionResetError,
        BrokenPipeError,
        EOFError,
    )


def _build_odbc_connnection_string(**kwargs: Any) -> str:
    return ";".join([f"{k}={v}" for k, v in kwargs.items()])


class SparkConnectionMethod(StrEnum):
    THRIFT = "thrift"
    HTTP = "http"
    ODBC = "odbc"
    SESSION = "session"


@dataclass
class SparkCredentials(Credentials):
    host: Optional[str] = None
    schema: Optional[str] = None  # type:ignore
    method: SparkConnectionMethod = None  # type: ignore
    database: Optional[str] = None  # type:ignore
    driver: Optional[str] = None
    cluster: Optional[str] = None
    endpoint: Optional[str] = None
    token: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    port: int = 443
    auth: Optional[str] = None
    kerberos_service_name: Optional[str] = None
    organization: str = "0"
    connection_string_suffix: Optional[str] = None
    connect_retries: int = 0
    connect_timeout: int = 10
    use_ssl: bool = False
    server_side_parameters: Dict[str, str] = field(default_factory=dict)
    retry_all: bool = False
    # Timeout in seconds for long-running queries (None = no timeout)
    # only used for PyhiveConnectionWrapper for now
    query_timeout: Optional[int] = None
    poll_interval: int = 5  # Polling interval in seconds for async queries
    query_retries: int = 1  # Number of times to retry on connection loss during query execution

    @classmethod
    def __pre_deserialize__(cls, data: Any) -> Any:
        data = super().__pre_deserialize__(data)
        if "database" not in data:
            data["database"] = None
        return data

    @property
    def cluster_id(self) -> Optional[str]:
        return self.cluster

    def __post_init__(self) -> None:
        if self.method is None:
            raise DbtRuntimeError("Must specify `method` in profile")
        if self.host is None:
            raise DbtRuntimeError("Must specify `host` in profile")
        if self.schema is None:
            raise DbtRuntimeError("Must specify `schema` in profile")

        # spark classifies database and schema as the same thing
        if self.database is not None and self.database != self.schema:
            raise DbtRuntimeError(
                f"    schema: {self.schema} \n"
                f"    database: {self.database} \n"
                f"On Spark, database must be omitted or have the same value as"
                f" schema."
            )
        self.database = None

        if self.method == SparkConnectionMethod.ODBC:
            try:
                import pyodbc  # noqa: F401
            except ImportError as e:
                raise DbtRuntimeError(
                    f"{self.method} connection method requires "
                    "additional dependencies. \n"
                    "Install the additional required dependencies with "
                    "`pip install dbt-spark[ODBC]`\n\n"
                    f"ImportError({e.msg})"
                ) from e

        if self.method == SparkConnectionMethod.ODBC and self.cluster and self.endpoint:
            raise DbtRuntimeError(
                "`cluster` and `endpoint` cannot both be set when"
                f" using {self.method} method to connect to Spark"
            )

        if (
            self.method == SparkConnectionMethod.HTTP
            or self.method == SparkConnectionMethod.THRIFT
        ) and not (ThriftState and THttpClient and hive):
            raise DbtRuntimeError(
                f"{self.method} connection method requires "
                "additional dependencies. \n"
                "Install the additional required dependencies with "
                "`pip install dbt-spark[PyHive]`"
            )

        if self.method == SparkConnectionMethod.SESSION:
            try:
                import pyspark  # noqa: F401
            except ImportError as e:
                raise DbtRuntimeError(
                    f"{self.method} connection method requires "
                    "additional dependencies. \n"
                    "Install the additional required dependencies with "
                    "`pip install dbt-spark[session]`\n\n"
                    f"ImportError({e.msg})"
                ) from e

        if self.method != SparkConnectionMethod.SESSION:
            self.host = self.host.rstrip("/")

        self.server_side_parameters = {
            str(key): str(value) for key, value in self.server_side_parameters.items()
        }

        # Disable ANSI mode to ensure compatibility with Spark v3 and v4
        # and prevent it from being overwritten.
        self.server_side_parameters["spark.sql.ansi.enabled"] = "false"

    @property
    def type(self) -> str:
        return "spark"

    @property
    def unique_field(self) -> str:
        return self.host  # type: ignore

    def _connection_keys(self) -> Tuple[str, ...]:
        return "host", "port", "cluster", "endpoint", "schema", "organization"


class SparkConnectionWrapper(ABC):
    @abstractmethod
    def cursor(self) -> "SparkConnectionWrapper":
        pass

    @abstractmethod
    def cancel(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def rollback(self) -> None:
        pass

    @abstractmethod
    def fetchall(self) -> Optional[List]:
        pass

    @abstractmethod
    def execute(self, sql: str, bindings: Optional[List[Any]] = None) -> None:
        pass

    @property
    @abstractmethod
    def description(
        self,
    ) -> Sequence[
        Tuple[str, Any, Optional[int], Optional[int], Optional[int], Optional[int], bool]
    ]:
        pass


class PyhiveConnectionWrapper(SparkConnectionWrapper):
    """Wrap a Spark connection in a way that no-ops transactions"""

    # https://forums.databricks.com/questions/2157/in-apache-spark-sql-can-we-roll-back-the-transacti.html  # noqa

    handle: "pyodbc.Connection"
    _cursor: "Optional[pyodbc.Cursor]"

    def __init__(
        self,
        handle: "pyodbc.Connection",
        poll_interval: int = 5,
        query_timeout: Optional[int] = None,
        query_retries: int = 1,
    ) -> None:
        self.handle = handle
        self._cursor = None
        self.poll_interval = poll_interval
        self.query_timeout = query_timeout
        self.query_retries = query_retries

    def cursor(self) -> "PyhiveConnectionWrapper":
        self._cursor = self.handle.cursor()
        return self

    def cancel(self) -> None:
        if self._cursor:
            # Handle bad response in the pyhive lib when
            # the connection is cancelled
            try:
                self._cursor.cancel()
            except EnvironmentError as exc:
                logger.debug("Exception while cancelling query: {}".format(exc))

    def close(self) -> None:
        if self._cursor:
            # Handle bad response in the pyhive lib when
            # the connection is cancelled
            try:
                self._cursor.close()
            except EnvironmentError as exc:
                logger.debug("Exception while closing cursor: {}".format(exc))
        self.handle.close()

    def rollback(self, *args: Any, **kwargs: Any) -> None:
        logger.debug("NotImplemented: rollback")

    def fetchall(self) -> List["pyodbc.Row"]:
        assert self._cursor, "Cursor not available"
        return self._cursor.fetchall()

    def execute(self, sql: str, bindings: Optional[List[Any]] = None) -> None:
        """Execute SQL with retries on connection loss during polling."""
        for attempt in range(self.query_retries + 1):
            try:
                self._execute_with_polling(sql, bindings)
                return  # Success - exit retry loop
            except Exception as e:
                if isinstance(e, CONNECTION_LOST_EXCEPTIONS):
                    # Connection was lost during polling
                    if attempt < self.query_retries:
                        logger.info(
                            f"Connection lost during query execution (attempt {attempt + 1}/{self.query_retries + 1}). "
                            f"Retrying in {self.poll_interval} seconds... "
                            f"Error: {type(e).__name__}: {str(e)}"
                        )
                        time.sleep(self.poll_interval)
                        # Need to get a fresh cursor on retry
                        self._cursor = self.handle.cursor()
                        continue
                    else:
                        # Exhausted retries - raise with helpful message
                        raise DbtRuntimeError(
                            f"Connection lost while polling query status after {self.query_retries + 1} attempts. "
                            "This often happens with long-running queries on idle connections. "
                            "The query may still be executing on the server. "
                            f"Original error: {type(e).__name__}: {str(e)}. "
                            f"Consider increasing 'query_retries' in your profile."
                        ) from e
                else:
                    # Not a connection exception - re-raise immediately
                    raise

    def _execute_with_polling(self, sql: str, bindings: Optional[List[Any]]) -> None:
        """Internal method that executes SQL and polls for completion."""
        if sql.strip().endswith(";"):
            sql = sql.strip()[:-1]

        # Reaching into the private enumeration here is bad form,
        # but there doesn't appear to be any way to determine that
        # a query has completed executing from the pyhive public API.
        # We need to use an async query + poll here, otherwise our
        # request may be dropped after ~5 minutes by the thrift server
        STATE_PENDING = [
            ThriftState.INITIALIZED_STATE,
            ThriftState.RUNNING_STATE,
            ThriftState.PENDING_STATE,
        ]

        STATE_SUCCESS = [
            ThriftState.FINISHED_STATE,
        ]

        if bindings is not None:
            bindings = [self._fix_binding(binding) for binding in bindings]

        assert self._cursor, "Cursor not available"

        self._cursor.execute(sql, bindings, async_=True)
        poll_state = self._cursor.poll()
        state = poll_state.operationState

        start_time = time.time()
        while state in STATE_PENDING:
            # Check for timeout if configured
            if self.query_timeout is not None:
                elapsed = time.time() - start_time
                if elapsed > self.query_timeout:
                    raise DbtRuntimeError(
                        f"Query exceeded timeout of {self.query_timeout} seconds"
                    )

            logger.debug("Poll status: {}, sleeping".format(state))
            time.sleep(self.poll_interval)

            # Poll and let connection exceptions bubble up
            # They will be caught and retried at the execute() method level
            poll_state = self._cursor.poll()
            state = poll_state.operationState

        # If an errorMessage is present, then raise a database exception
        # with that exact message. If no errorMessage is present, the
        # query did not necessarily succeed: check the state against the
        # known successful states, raising an error if the query did not
        # complete in a known good state. This can happen when queries are
        # cancelled, for instance. The errorMessage will be None, but the
        # state of the query will be "cancelled". By raising an exception
        # here, we prevent dbt from showing a status of OK when the query
        # has in fact failed.
        if poll_state.errorMessage:
            logger.debug("Poll response: {}".format(poll_state))
            logger.debug("Poll status: {}".format(state))
            raise DbtDatabaseError(poll_state.errorMessage)

        elif state not in STATE_SUCCESS:
            status_type = ThriftState._VALUES_TO_NAMES.get(state, "Unknown<{!r}>".format(state))
            raise DbtDatabaseError("Query failed with status: {}".format(status_type))

        logger.debug("Poll status: {}, query complete".format(state))

    @classmethod
    def _fix_binding(cls, value: Any) -> Union[float, str]:
        """Convert complex datatypes to primitives that can be loaded by
        the Spark driver"""
        if isinstance(value, NUMBERS):
            return float(value)
        elif isinstance(value, datetime):
            return value.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        else:
            return value

    @property
    def description(
        self,
    ) -> Sequence[
        Tuple[str, Any, Optional[int], Optional[int], Optional[int], Optional[int], bool]
    ]:
        assert self._cursor, "Cursor not available"
        return self._cursor.description


class PyodbcConnectionWrapper(PyhiveConnectionWrapper):
    def execute(self, sql: str, bindings: Optional[List[Any]] = None) -> None:
        assert self._cursor, "Cursor not available"
        if sql.strip().endswith(";"):
            sql = sql.strip()[:-1]
        # pyodbc does not handle a None type binding!
        if bindings is None:
            self._cursor.execute(sql)
        else:
            # pyodbc only supports `qmark` sql params!
            query = sqlparams.SQLParams("format", "qmark")
            sql, bindings = query.format(sql, bindings)
            self._cursor.execute(sql, *bindings)


class SparkConnectionManager(SQLConnectionManager):
    TYPE = "spark"

    SPARK_CLUSTER_HTTP_PATH = "/sql/protocolv1/o/{organization}/{cluster}"
    SPARK_SQL_ENDPOINT_HTTP_PATH = "/sql/1.0/endpoints/{endpoint}"
    SPARK_CONNECTION_URL = "{host}:{port}" + SPARK_CLUSTER_HTTP_PATH

    @contextmanager
    def exception_handler(self, sql: str) -> Generator[None, None, None]:
        try:
            yield

        except Exception as exc:
            logger.debug("Error while running:\n{}".format(sql))
            logger.debug(exc)
            if len(exc.args) == 0:
                raise

            thrift_resp = exc.args[0]
            if hasattr(thrift_resp, "status"):
                msg = thrift_resp.status.errorMessage
                raise DbtRuntimeError(msg)
            else:
                raise DbtRuntimeError(str(exc))

    def cancel(self, connection: Connection) -> None:
        connection.handle.cancel()

    @classmethod
    def get_response(cls, cursor: Any) -> AdapterResponse:
        # https://github.com/dbt-labs/dbt-spark/issues/142
        message = "OK"
        return AdapterResponse(_message=message)

    # No transactions on Spark....
    def add_begin_query(self, *args: Any, **kwargs: Any) -> None:
        logger.debug("NotImplemented: add_begin_query")

    def add_commit_query(self, *args: Any, **kwargs: Any) -> None:
        logger.debug("NotImplemented: add_commit_query")

    def commit(self, *args: Any, **kwargs: Any) -> None:
        logger.debug("NotImplemented: commit")

    def rollback(self, *args: Any, **kwargs: Any) -> None:
        logger.debug("NotImplemented: rollback")

    @classmethod
    def validate_creds(cls, creds: Any, required: Iterable[str]) -> None:
        method = creds.method

        for key in required:
            if not hasattr(creds, key):
                raise DbtConfigError(
                    "The config '{}' is required when using the {} method"
                    " to connect to Spark".format(key, method)
                )

    @classmethod
    def open(cls, connection: Connection) -> Connection:
        if connection.state == ConnectionState.OPEN:
            logger.debug("Connection is already open, skipping open.")
            return connection

        creds = connection.credentials
        exc = None
        handle: SparkConnectionWrapper

        for i in range(1 + creds.connect_retries):
            try:
                if creds.method == SparkConnectionMethod.HTTP:
                    cls.validate_creds(creds, ["token", "host", "port", "cluster", "organization"])

                    # Prepend https:// if it is missing
                    host = creds.host
                    if not host.startswith("https://"):
                        host = "https://" + creds.host

                    conn_url = cls.SPARK_CONNECTION_URL.format(
                        host=host,
                        port=creds.port,
                        organization=creds.organization,
                        cluster=creds.cluster,
                    )

                    logger.debug("connection url: {}".format(conn_url))

                    transport = THttpClient.THttpClient(conn_url)

                    raw_token = "token:{}".format(creds.token).encode()
                    token = base64.standard_b64encode(raw_token).decode()
                    transport.setCustomHeaders({"Authorization": "Basic {}".format(token)})

                    conn = hive.connect(
                        thrift_transport=transport,
                        configuration=creds.server_side_parameters,
                    )
                    handle = PyhiveConnectionWrapper(
                        conn,
                        poll_interval=creds.poll_interval,
                        query_timeout=creds.query_timeout,
                        query_retries=creds.query_retries,
                    )
                elif creds.method == SparkConnectionMethod.THRIFT:
                    cls.validate_creds(creds, ["host", "port", "user", "schema"])

                    if creds.use_ssl:
                        transport = build_ssl_transport(
                            host=creds.host,
                            port=creds.port,
                            username=creds.user,
                            auth=creds.auth,
                            kerberos_service_name=creds.kerberos_service_name,
                            password=creds.password,
                        )
                        conn = hive.connect(
                            thrift_transport=transport,
                            configuration=creds.server_side_parameters,
                        )
                    else:
                        conn = hive.connect(
                            host=creds.host,
                            port=creds.port,
                            username=creds.user,
                            auth=creds.auth,
                            kerberos_service_name=creds.kerberos_service_name,
                            password=creds.password,
                            configuration=creds.server_side_parameters,
                        )  # noqa
                    handle = PyhiveConnectionWrapper(
                        conn,
                        poll_interval=creds.poll_interval,
                        query_timeout=creds.query_timeout,
                        query_retries=creds.query_retries,
                    )
                elif creds.method == SparkConnectionMethod.ODBC:
                    if creds.cluster is not None:
                        required_fields = [
                            "driver",
                            "host",
                            "port",
                            "token",
                            "organization",
                            "cluster",
                        ]
                        http_path = cls.SPARK_CLUSTER_HTTP_PATH.format(
                            organization=creds.organization, cluster=creds.cluster
                        )
                    elif creds.endpoint is not None:
                        required_fields = ["driver", "host", "port", "token", "endpoint"]
                        http_path = cls.SPARK_SQL_ENDPOINT_HTTP_PATH.format(
                            endpoint=creds.endpoint
                        )
                    elif creds.connection_string_suffix is not None:
                        required_fields = ["driver", "host", "port", "connection_string_suffix"]
                    else:
                        raise DbtConfigError(
                            "Either `cluster`, `endpoint`, `connection_string_suffix` must set when"
                            " using the odbc method to connect to Spark"
                        )

                    cls.validate_creds(creds, required_fields)
                    dbt_spark_version = __version__.version
                    user_agent_entry = (
                        f"dbt-labs-dbt-spark/{dbt_spark_version} (Databricks)"  # noqa
                    )
                    # http://simba.wpengine.com/products/Spark/doc/ODBC_InstallGuide/unix/content/odbc/hi/configuring/serverside.htm
                    ssp = {f"SSP_{k}": f"{{{v}}}" for k, v in creds.server_side_parameters.items()}
                    if creds.token is not None:
                        # https://www.simba.com/products/Spark/doc/v2/ODBC_InstallGuide/unix/content/odbc/options/driver.htm
                        connection_str = _build_odbc_connnection_string(
                            DRIVER=creds.driver,
                            HOST=creds.host,
                            PORT=creds.port,
                            UID="token",
                            PWD=creds.token,
                            HTTPPath=http_path,
                            AuthMech=3,
                            SparkServerType=3,
                            ThriftTransport=2,
                            SSL=1,
                            UserAgentEntry=user_agent_entry,
                            LCaseSspKeyName=0 if ssp else 1,
                            **ssp,
                        )
                    else:
                        connection_str = _build_odbc_connnection_string(
                            DRIVER=creds.driver,
                            HOST=creds.host,
                            PORT=creds.port,
                            ThriftTransport=2,
                            SSL=1,
                            UserAgentEntry=user_agent_entry,
                            LCaseSspKeyName=0 if ssp else 1,
                            **ssp,
                        )
                    if creds.connection_string_suffix is not None:
                        connection_str = connection_str + ";" + creds.connection_string_suffix

                    conn = pyodbc.connect(connection_str, autocommit=True)
                    handle = PyodbcConnectionWrapper(conn)
                elif creds.method == SparkConnectionMethod.SESSION:
                    from .session import (  # noqa: F401
                        Connection,
                        SessionConnectionWrapper,
                    )

                    handle = SessionConnectionWrapper(
                        Connection(server_side_parameters=creds.server_side_parameters)
                    )
                else:
                    raise DbtConfigError(f"invalid credential method: {creds.method}")
                break
            except Exception as e:
                exc = e
                if isinstance(e, EOFError):
                    # The user almost certainly has invalid credentials.
                    # Perhaps a token expired, or something
                    msg = "Failed to connect"
                    if creds.token is not None:
                        msg += ", is your token valid?"
                    raise FailedToConnectError(msg) from e
                retryable_message = _is_retryable_error(e)
                if retryable_message and creds.connect_retries > 0:
                    msg = (
                        f"Warning: {retryable_message}\n\tRetrying in "
                        f"{creds.connect_timeout} seconds "
                        f"({i} of {creds.connect_retries})"
                    )
                    logger.warning(msg)
                    time.sleep(creds.connect_timeout)
                elif creds.retry_all and creds.connect_retries > 0:
                    msg = (
                        f"Warning: {getattr(exc, 'message', 'No message')}, "
                        f"retrying due to 'retry_all' configuration "
                        f"set to true.\n\tRetrying in "
                        f"{creds.connect_timeout} seconds "
                        f"({i} of {creds.connect_retries})"
                    )
                    logger.warning(msg)
                    time.sleep(creds.connect_timeout)
                else:
                    raise FailedToConnectError("failed to connect") from e
        else:
            raise exc  # type: ignore

        connection.handle = handle
        connection.state = ConnectionState.OPEN  # type:ignore
        return connection

    @classmethod
    def data_type_code_to_name(cls, type_code: Union[type, str]) -> str:  # type:ignore
        """
        :param Union[type, str] type_code: The sql to execute.
            * type_code is a python type (!) in pyodbc https://github.com/mkleehammer/pyodbc/wiki/Cursor#description, and a string for other spark runtimes.
            * ignoring the type annotation on the signature for this adapter instead of updating the base class because this feels like a really special case.
        :return: stringified the cursor type_code
        :rtype: str
        """
        if isinstance(type_code, str):
            return type_code
        return type_code.__name__.upper()


def build_ssl_transport(
    host: str,
    port: int,
    username: str,
    auth: str,
    kerberos_service_name: str,
    password: Optional[str] = None,
) -> "thrift_sasl.TSaslClientTransport":
    transport = None
    if port is None:
        port = 10000
    if auth is None:
        auth = "NONE"
    socket = TSSLSocket(host, port, cert_reqs=ssl.CERT_NONE)
    if auth == "NOSASL":
        # NOSASL corresponds to hive.server2.authentication=NOSASL
        # in hive-site.xml
        transport = thrift.transport.TTransport.TBufferedTransport(socket)
    elif auth in ("LDAP", "KERBEROS", "NONE", "CUSTOM"):
        # Defer import so package dependency is optional
        if auth == "KERBEROS":
            # KERBEROS mode in hive.server2.authentication is GSSAPI
            # in sasl library
            sasl_auth = "GSSAPI"
        else:
            sasl_auth = "PLAIN"
            if password is None:
                # Password doesn't matter in NONE mode, just needs
                # to be nonempty.
                password = "x"

        def sasl_factory() -> SASLClient:
            if sasl_auth == "GSSAPI":
                sasl_client = SASLClient(host, kerberos_service_name, mechanism=sasl_auth)
            elif sasl_auth == "PLAIN":
                sasl_client = SASLClient(
                    host, mechanism=sasl_auth, username=username, password=password
                )
            else:
                raise AssertionError
            return sasl_client

        transport = thrift_sasl.TSaslClientTransport(sasl_factory, sasl_auth, socket)
    return transport


def _is_retryable_error(exc: Exception) -> str:
    message = str(exc).lower()
    if "pending" in message or "temporarily_unavailable" in message:
        return str(exc)
    else:
        return ""



import pytest

from dbt.tests.adapter.incremental.test_incremental_microbatch import (
    BaseMicrobatch,
)

# No requirement for a unique_id for spark microbatch!
_microbatch_model_no_unique_id_sql = """
{{ config(materialized='incremental', incremental_strategy='microbatch', event_time='event_time', batch_size='day', begin=modules.datetime.datetime(2020, 1, 1, 0, 0, 0), partition_by=['date_day'], file_format='parquet') }}
select *, cast(event_time as date) as date_day
from {{ ref('input_model') }}
"""


@pytest.mark.skip_profile(
    "databricks_http_cluster", "databricks_sql_endpoint", "spark_session", "spark_http_odbc"
)
class TestMicrobatch(BaseMicrobatch):
    @pytest.fixture(scope="class")
    def microbatch_model_sql(self) -> str:
        return _microbatch_model_no_unique_id_sql



import pytest
from dbt.tests.adapter.incremental.test_incremental_unique_id import BaseIncrementalUniqueKey


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestUniqueKeySpark(BaseIncrementalUniqueKey):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
                "+incremental_strategy": "merge",
            }
        }



import pytest
from dbt.tests.adapter.utils.data_types.test_type_bigint import BaseTypeBigInt
from dbt.tests.adapter.utils.data_types.test_type_float import (
    BaseTypeFloat,
    seeds__expected_csv as seeds__float_expected_csv,
)
from dbt.tests.adapter.utils.data_types.test_type_int import (
    BaseTypeInt,
    seeds__expected_csv as seeds__int_expected_csv,
)
from dbt.tests.adapter.utils.data_types.test_type_numeric import BaseTypeNumeric
from dbt.tests.adapter.utils.data_types.test_type_string import BaseTypeString
from dbt.tests.adapter.utils.data_types.test_type_timestamp import BaseTypeTimestamp
from dbt.tests.adapter.utils.data_types.test_type_boolean import BaseTypeBoolean


class TestTypeBigInt(BaseTypeBigInt):
    pass


# need to explicitly cast this to avoid it being inferred/loaded as a DOUBLE on Spark
# in SparkSQL, the two are equivalent for `=` comparison, but distinct for EXCEPT comparison
seeds__float_expected_yml = """
version: 2
seeds:
  - name: expected
    config:
      column_types:
        float_col: float
"""


class TestTypeFloat(BaseTypeFloat):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "expected.csv": seeds__float_expected_csv,
            "expected.yml": seeds__float_expected_yml,
        }


# need to explicitly cast this to avoid it being inferred/loaded as a BIGINT on Spark
seeds__int_expected_yml = """
version: 2
seeds:
  - name: expected
    config:
      column_types:
        int_col: int
"""


class TestTypeInt(BaseTypeInt):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "expected.csv": seeds__int_expected_csv,
            "expected.yml": seeds__int_expected_yml,
        }


class TestTypeNumeric(BaseTypeNumeric):
    def numeric_fixture_type(self):
        return "decimal(28,6)"


class TestTypeString(BaseTypeString):
    pass


class TestTypeTimestamp(BaseTypeTimestamp):
    pass


class TestTypeBoolean(BaseTypeBoolean):
    pass
